from keras import Model, ops
from keras.callbacks import Callback

from ...layers import QLayerBase


class FreeEBOPs(Callback):
    def on_epoch_end(self, epoch, logs=None):
        assert logs is not None
        assert isinstance(self.model, Model)
        ebops = 0
        for layer in self.model._flatten_layers():
            if isinstance(layer, QLayerBase) and layer.enable_ebops:
                ebops += int(ops.convert_to_numpy(layer._ebops))  # type: ignore
        logs['ebops'] = ebops



import inspect
import re
from typing import Any

import numpy as np

numbers = int | float | np.integer | np.floating


def gather_vars_to_kwargs(skip_pattern=None) -> dict[str, Any]:
    """Gather all local variables in the calling function and return them as a dictionary. If a variable is named `kwargs`, it will be updated with the rest of the variables. This function is useful for initializing classes with a large number of arguments.

    Parameters
    ----------
    skip_pattern : str, optional
        A regular expression pattern to match variables to skip. If a variable matches the pattern, it will not be included in the returned dictionary. The default is None.

    Returns
    -------
    dict[str, Any]
        A dictionary containing all local variables in the calling function, except for those that start and end with double underscores.
    """
    vars = dict(inspect.getouterframes(inspect.currentframe(), 2)[1][0].f_locals)
    kwarg = vars.pop('kwargs', {})
    if skip_pattern:
        m = re.compile(skip_pattern)
        for k in list(vars.keys()):
            if m.match(k):
                del vars[k]
    kwarg.update(vars)
    for k in list(kwarg.keys()):
        if k.startswith('__') and k.endswith('__'):
            del kwarg[k]
    return kwarg



import keras
from keras import ops
from keras.constraints import Constraint
from keras.initializers import Constant, Initializer
from keras.regularizers import Regularizer
from quantizers.fixed_point._fixed_point_ops import round_conv
from quantizers.minifloat.float_point_ops import float_quantize

from ...constraints import Min
from .base import TrainableQuantizerBase, numbers


class FloatPointQuantizer(TrainableQuantizerBase):
    """Internal class for float-point quantizer. Follows IEEE 754 standard with subnormal numbers, but without NaNs and infinities. The quantizer is defined by three parameters: mantissa, exponent, and exponent offset. Mantissa bits are excluding the sign bit. Exponent bits are including the sign bit. Exponent offset is added to the signed exponent.

    The sign bit always presents in this quantizer. However, as the number of mantissa bits reaches -1, the quantizer will always produce zero.

    Can be used as a quantizer in Keras layers, but is usually wrapped by a `Quantizer` class to provide a consistent interface.

    Parameters
    ----------
    m0 : numbers | Initializer
        Initial value of the number of mantissa bits. Trainable.
    e0 : numbers | Initializer
        Initial value of the number of exponent bits. Trainable.
    e00 : numbers | Initializer, optional
        Initial value of the exponent offset. Default is 0. Trainable.
    mc : Constraint, optional
        Constraint for the number of mantissa bits. Default is Min(-1).
    ec : Constraint, optional
        Constraint for the number of exponent bits. Default is keras.constraints.NonNeg().
    e0c : Constraint, optional
        Constraint for the exponent offset. Default is None.
    mr : Regularizer, optional
        Regularizer for the number of mantissa bits. Default is None.
    er : Regularizer, optional
        Regularizer for the number of exponent bits. Default is None.
    e0r : Regularizer, optional
        Regularizer for the exponent offset. Default is None.
    """

    def __init__(
        self,
        m0: numbers | Initializer,
        e0: numbers | Initializer,
        e00: numbers | Initializer = 0,
        mc: Constraint | None = Min(-1),
        ec: Constraint | None = keras.constraints.NonNeg(),
        e0c: Constraint | None = None,
        mr: Regularizer | None = None,
        er: Regularizer | None = None,
        e0r: Regularizer | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._m0 = m0 if isinstance(m0, Initializer) else Constant(float(m0))
        self._e0 = e0 if isinstance(e0, Initializer) else Constant(float(e0))
        self._e00 = e00 if isinstance(e00, Initializer) else Constant(float(e00))
        self.m_constraint = mc
        self.e_constraint = ec
        self.e0_constraint = e0c
        self.m_regularizer = mr
        self.e_regularizer = er
        self.e0_regularizer = e0r

    def build(self, input_shape):
        super().build(input_shape)
        bw_shape = self.bw_mapper.inference_weight_shape(input_shape)
        self._m = self.add_weight(
            name='m',
            shape=bw_shape,
            initializer=self._m0,
            trainable=True,
            constraint=self.m_constraint,
            regularizer=self.m_regularizer,
        )
        self._e = self.add_weight(
            name='e',
            shape=bw_shape,
            initializer=self._e0,
            trainable=True,
            constraint=self.e_constraint,
            regularizer=self.e_regularizer,
        )
        self._e0 = self.add_weight(
            name='e0',
            shape=bw_shape,
            initializer=self._e00,
            trainable=True,
            constraint=self.e0_constraint,
            regularizer=self.e0_regularizer,
        )
        super().build(input_shape)

    @property
    def m(self):
        return round_conv(ops.cast(self._m, ops.dtype(self._m)))  # type: ignore

    @property
    def e(self):
        return round_conv(ops.cast(self._e, ops.dtype(self._e)))  # type: ignore

    @property
    def e0(self):
        return round_conv(ops.cast(self._e0, ops.dtype(self._e0)))  # type: ignore

    @property
    def bits(self):
        return self.m + self.e + 1.0  # type: ignore

    @property
    def fbits(self):
        return self.bits

    @property
    def min(self):
        return -self.max

    @property
    def max(self):
        return 2.0 ** (2 ** (self.e - 1.0) + self.e0 - 1) * (2 - 2.0**-self.m)  # type: ignore

    @property
    def epsilon(self):
        return 2.0 ** (-(2 ** (self.e - 1)) + self.e0 + 1) * (2.0**-self.m)  # type: ignore

    def call(self, inputs, training=None):
        m = self.bw_mapper.bw_to_x(self.m, ops.shape(inputs))
        e = self.bw_mapper.bw_to_x(self.e, ops.shape(inputs))
        e0 = self.bw_mapper.bw_to_x(self.e0, ops.shape(inputs))
        return float_quantize(inputs, m, e, e0)

    def __repr__(self):
        if not self.built:
            return f'{self.__class__.__name__}(name={self.name}, built=False)'
        mstd, estd, e0std = float(ops.std(self.m)), float(ops.std(self.e)), float(ops.std(self.e0))  # type: ignore
        mmean, emean, e0mean = float(ops.mean(self.m)), float(ops.mean(self.e)), float(ops.mean(self.e0))  # type: ignore
        mstr = f'{mmean:.2f}±{mstd:.2f}'
        estr = f'{emean:.2f}±{estd:.2f}'
        e0str = f'{e0mean:.2f}±{e0std:.2f}'
        return f'{self.__class__.__name__}(m={mstr}, e={estr}, e0={e0str}, name={self.name})'



from typing import Literal

from keras import ops

from ..quantizer.config import QuantizerConfig
from ..utils.misc import gather_vars_to_kwargs
from .core.einsum_dense import QEinsumDense
from .multi_head_attention import QMultiHeadAttention


class QLinformerAttention(QMultiHeadAttention):
    __output_quantizer_handled__ = True

    def __init__(
        self,
        num_heads,
        lin_kv_proj_dim,
        key_dim,
        value_dim=None,
        dropout=0.0,
        use_bias=True,
        output_shape=None,
        attention_axes=None,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        seed=None,
        fuse: Literal['none', 'qkv', 'kv'] = 'none',
        qkvo_iq_conf: QuantizerConfig | None = None,
        qkvo_kq_conf: QuantizerConfig | None = None,
        qkvo_bq_conf: QuantizerConfig | None = None,
        qkvo_oq_conf: QuantizerConfig | None = None,
        softmax_iq_conf: QuantizerConfig | None = None,
        softmax_exp_iq_conf: QuantizerConfig | None = None,
        softmax_exp_oq_conf: QuantizerConfig | None = None,
        softmax_inv_iq_conf: QuantizerConfig | None = None,
        softmax_inv_oq_conf: QuantizerConfig | None = None,
        softmax_oq_conf: QuantizerConfig | None = None,
        stable_softmax=True,
        softmax_allow_heterogeneous_table: bool = False,
        parallelization_factor=-1,
        **kwargs,
    ):
        kwargs = gather_vars_to_kwargs('self|lin_kv_proj_dim')
        self._kv_proj_dim = (lin_kv_proj_dim,) if isinstance(lin_kv_proj_dim, int) else tuple(lin_kv_proj_dim)
        super().__init__(**kwargs)

    def build(self, query_shape, value_shape, key_shape=None):
        key_shape = key_shape or value_shape
        key_rank = len(key_shape)
        value_rank = len(value_shape)
        assert key_rank == value_rank, (
            f'Key and value must have the same rank, but got key shape {key_shape} and value shape {value_shape}.'
        )
        if self.attention_axes is not None:
            attn_axes = tuple(self.attention_axes)
        else:
            attn_axes = tuple(range(1, value_rank - 1))

        assert len(attn_axes) == len(self._kv_proj_dim), (
            f'Attention axes are {attn_axes}, but kv_proj_dim is {self._kv_proj_dim}. They must match in length.'
        )

        _value_shape_proj, _key_shape_proj = list(value_shape), list(key_shape)
        for i, j in enumerate(attn_axes):
            _value_shape_proj[j] = self._kv_proj_dim[i]
            _key_shape_proj[j] = self._kv_proj_dim[i]
        self._key_shape_proj = tuple(_key_shape_proj)
        self._value_shape_proj = tuple(_value_shape_proj)

        template = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
        inp_idx = template[:value_rank]
        attn_idx = ''.join(inp_idx[i] for i in attn_axes)
        new_attn_idx = template[value_rank : value_rank + len(self._kv_proj_dim)]
        ker_idx = attn_idx + new_attn_idx
        _out_idx = list(inp_idx)
        for i, j in enumerate(attn_axes):
            _out_idx[j] = new_attn_idx[i]
        out_idx = ''.join(_out_idx)
        eq_lin_kv_proj = f'{inp_idx},{ker_idx}->{out_idx}'

        self._lin_k_proj = QEinsumDense(
            eq_lin_kv_proj, self._key_shape_proj[1:], bias_axes=None, **self._get_common_kwargs_for_sublayer()
        )

        self._lin_v_proj = QEinsumDense(
            eq_lin_kv_proj, self._value_shape_proj[1:], bias_axes=None, **self._get_common_kwargs_for_sublayer()
        )

        self._lin_k_proj.build(key_shape)
        self._lin_v_proj.build(value_shape)

        super().build(query_shape, self._value_shape_proj, key_shape=self._key_shape_proj)

    def call(
        self,
        query,
        value,
        key=None,
        query_mask=None,
        value_mask=None,
        key_mask=None,
        attention_mask=None,
        return_attention_scores=False,
        training=None,
        use_causal_mask=False,
    ):
        assert use_causal_mask is False, 'Causal mask is not supported in QLinformerAttention.'
        key = key if key is not None else value
        key = self._lin_k_proj(key, training=training)
        value = self._lin_v_proj(value, training=training)
        return super().call(
            query,
            value,
            key,
            query_mask=query_mask,
            value_mask=value_mask,
            key_mask=key_mask,
            attention_mask=attention_mask,
            return_attention_scores=return_attention_scores,
            training=training,
        )

    @property
    def ebops(self) -> int:
        if self._ebops is None:
            return 0
        ebops = sum(
            (  # type: ignore
                self._query_dense.ebops,
                self._key_dense.ebops,
                self._value_dense.ebops,
                self._softmax.ebops,
                self._output_dense.ebops,
                self._lin_k_proj.ebops,
                self._lin_v_proj.ebops,
                ops.convert_to_tensor(self._ebops),
            )
        )
        return round(ops.convert_to_numpy(ebops).item())  # type: ignore

    def _compute_ebops(self, query_shape, value_shape, key_shape=None):
        return super()._compute_ebops(query_shape, self._value_shape_proj, key_shape=self._key_shape_proj)

    def get_config(self):
        config = super().get_config()
        config['lin_kv_proj_dim'] = self._kv_proj_dim
        return config



from ..quantizer import Quantizer
from .activation import QUnaryFunctionLUT
from .batch_normalization import QBatchNormalization
from .conv import QConv1D, QConv2D, QConv3D
from .core import *
from .einsum_dense_batchnorm import QEinsumDenseBatchnorm
from .linformer_attention import QLinformerAttention
from .multi_head_attention import QMultiHeadAttention
from .ops import *
from .pooling import *
from .pooling import (
    QAveragePooling1D as QAvgPool1D,
)
from .pooling import (
    QAveragePooling2D as QAvgPool2D,
)
from .pooling import (
    QAveragePooling3D as QAvgPool3D,
)
from .pooling import (
    QGlobalAveragePooling1D as QGlobalAvgPool1D,
)
from .pooling import (
    QGlobalAveragePooling2D as QGlobalAvgPool2D,
)
from .pooling import (
    QGlobalAveragePooling3D as QGlobalAvgPool3D,
)
from .pooling import (
    QGlobalMaxPooling1D as QGlobalMaxPool1D,
)
from .pooling import (
    QGlobalMaxPooling2D as QGlobalMaxPool2D,
)
from .pooling import (
    QGlobalMaxPooling3D as QGlobalMaxPool3D,
)
from .pooling import (
    QMaxPooling1D as QMaxPool1D,
)
from .pooling import (
    QMaxPooling2D as QMaxPool2D,
)
from .pooling import (
    QMaxPooling3D as QMaxPool3D,
)
from .rnn import QGRU, QSimpleRNN
from .softmax import QSoftmax
from .table import QDenseT

__all__ = [
    'QUnaryFunctionLUT',
    'QBatchNormalization',
    'QConv1D',
    'QConv2D',
    'QConv3D',
    'QEinsumDenseBatchnorm',
    'QSoftmax',
    'Quantizer',
    'QAdd',
    'QDot',
    'QEinsumDense',
    'QMeanPow2',
    'QSum',
    'QAdd',
    'QAveragePow2',
    'QDot',
    'QEinsum',
    'QMaximum',
    'QMinimum',
    'QMultiply',
    'QSubtract',
    'QLinformerAttention',
    'QMultiHeadAttention',
    'QBatchNormDense',
    'QDense',
    'QMaxPool1D',
    'QMaxPool2D',
    'QMaxPool3D',
    'QAvgPool1D',
    'QAvgPool2D',
    'QAvgPool3D',
    'QGlobalAvgPool1D',
    'QGlobalAvgPool2D',
    'QGlobalAvgPool3D',
    'QGlobalMaxPool1D',
    'QGlobalMaxPool2D',
    'QGlobalMaxPool3D',
    'QMaxPooling1D',
    'QMaxPooling2D',
    'QMaxPooling3D',
    'QAveragePooling1D',
    'QAveragePooling2D',
    'QAveragePooling3D',
    'QGlobalAveragePooling1D',
    'QGlobalAveragePooling2D',
    'QGlobalAveragePooling3D',
    'QGlobalMaxPooling1D',
    'QGlobalMaxPooling2D',
    'QGlobalMaxPooling3D',
    'QSimpleRNN',
    'QGRU',
    'QDenseT',
]



from typing import TypedDict, overload

from keras.initializers import Initializer


class GlobalConfig(TypedDict):
    beta0: float
    enable_ebops: bool
    enable_oq: bool
    enable_iq: bool


global_config = GlobalConfig(
    beta0=1e-5,
    enable_ebops=True,
    enable_oq=False,
    enable_iq=True,
)


class LayerConfigScope:
    @overload
    def __init__(self, *, beta0: float | None | Initializer = None, enable_ebops: bool | None = None): ...

    @overload
    def __init__(self, **kwargs): ...

    def __init__(self, **kwargs):
        self._override = kwargs

    def __enter__(self):
        self._tmp = global_config.copy()
        for k, v in self._override.items():
            global_config[k] = v

    def __exit__(self, exc_type, exc_val, exc_tb):
        for k in self._override:
            del global_config[k]
        global_config.update(self._tmp)



import inspect
from abc import ABCMeta
from collections.abc import Callable, Iterable, Sequence
from functools import wraps

import numpy as np
from keras import ops
from keras.initializers import Constant, Initializer
from keras.layers import Layer
from keras.saving import deserialize_keras_object, register_keras_serializable, serialize_keras_object

from ...config.layer import global_config
from ...quantizer import Quantizer, QuantizerConfig


def get_method_source(cls, method_name):
    "Get the class that defines the method"
    for parent_cls in inspect.getmro(cls):
        if method_name in parent_cls.__dict__:
            return parent_cls
    assert False, f'Method {method_name} is not defined in {cls}'


def check_save_load_own_variables(cls):
    "Assert that save_own_variables and load_own_variables are not overriden, or they should be defined in the same class"

    if cls.save_own_variables is not Layer.save_own_variables:
        cls_save = get_method_source(cls, 'save_own_variables')
        assert cls_save is cls, f'save_own_variables in {cls} is overriden in {cls_save}, which will likely break save_weights fn'
    if cls.load_own_variables is not Layer.load_own_variables:
        cls_load = get_method_source(cls, 'load_own_variables')
        assert cls_load is cls, f'load_own_variables in {cls} is overriden in {cls_load}, which will likely break load_weights fn'


class QLayerMeta(ABCMeta):
    _wrapped_cls = set()

    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)
        check_save_load_own_variables(cls)
        # ====================================================================
        # =========== Register as Keras serializable if possible =============
        # ====================================================================

        if cls.get_config is not Layer.get_config and cls.__module__.startswith('hgq'):
            original_get_config = cls.get_config

            @wraps(original_get_config)
            def get_config(self):
                config = original_get_config(self)
                config = serialize_keras_object(config)
                return config

            cls.get_config = get_config  # type: ignore
            cls = register_keras_serializable(package='hgq')(cls)

    def __call__(cls: type, *args, **kwargs):
        if cls in QLayerMeta._wrapped_cls:
            return super().__call__(*args, **kwargs)  # type: ignore
        QLayerMeta._wrapped_cls.add(cls)

        # ====================================================================
        # ============ Compute ebops if _compute_ebops presents ==============
        # =============== Apply output quantizer if possible =================
        # ====================================================================

        original_call: Callable = cls.call
        _compute_ebops = cls._compute_ebops

        if original_call is not Layer.call and _compute_ebops is not QLayerBase._compute_ebops:
            VAR_KEYWORD = inspect.Parameter.VAR_KEYWORD
            KEYWORD_ONLY = inspect.Parameter.KEYWORD_ONLY
            signature = inspect.signature(original_call)

            # Add kwarg training to signature if not present
            has_training = 'training' in signature.parameters
            has_var_keyword = any(v.kind == VAR_KEYWORD for v in signature.parameters.values())
            new_signature = signature

            if not has_training and not has_var_keyword:
                training_param = inspect.Parameter('training', KEYWORD_ONLY, default=None)
                new_params = signature.parameters.copy()
                new_params['training'] = training_param
                new_signature = signature.replace(parameters=new_params.values())  # type: ignore

            if not getattr(cls, '__no_wrap_call__', False):

                @wraps(original_call)
                def call(self, *args, **kwargs):
                    if has_training:  # remove training flag if the original does not have it
                        training = kwargs.get('training', None)
                    else:
                        training = kwargs.pop('training', None)
                    r = original_call(self, *args, **kwargs)
                    if training or training == 'tracing' and self.enable_ebops:  # noqa: E712, training may be a special wrapper object
                        if isinstance(args[0], (tuple, list)):
                            tensors = args[0]
                        else:
                            tensors = args
                        shapes = ((1,) + shape[1:] for shape in map(ops.shape, tensors))
                        if self.enable_ebops:
                            ebops = self._compute_ebops(*shapes)
                            self._ebops.assign(ops.cast(ebops, self._ebops.dtype))
                            self.add_loss(ebops * self.beta)
                    if not self.enable_oq or self.__output_quantizer_handled__:
                        return r
                    assert not isinstance(r, (tuple, list)), (
                        f'Layer {self.name}({type(self)}) returns multiple outputs, which must be handled in subclasses.'
                    )
                    return self.oq(r, training=training)

                call.__signature__ = new_signature  # type: ignore

                cls.call = call

        return super().__call__(*args, **kwargs)  # type: ignore


class QLayerBase(Layer, metaclass=QLayerMeta):
    """Base class for quantized layers.

    Handles: instantiation of input and output quantizers,
    ebops and beta variable creation.

    call method is wrapped to compute ebops and apply output quantizer if applicable.
    """

    save_own_variables = Layer.save_own_variables
    load_own_variables = Layer.load_own_variables
    __output_quantizer_handled__ = False

    def __init__(
        self,
        enable_ebops: bool | None = None,
        beta0: int | float | np.number | None | Initializer = None,
        enable_oq: bool | None = None,
        enable_iq: bool | None = None,
        oq_conf: QuantizerConfig | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        beta0 = beta0 if beta0 is not None else global_config['beta0']
        beta0 = Constant(float(beta0)) if not isinstance(beta0, Initializer) else beta0

        self._enable_iq = enable_iq if enable_iq is not None else global_config['enable_iq']
        self._enable_oq = enable_oq if enable_oq is not None else global_config['enable_oq']
        if enable_ebops is None:
            enable_ebops = global_config['enable_ebops'] and self.enable_iq
        self._enable_ebops = enable_ebops
        self._beta0 = beta0

        if self.enable_oq:
            oq_conf = oq_conf or QuantizerConfig('default', 'datalane')
            self._oq = Quantizer(oq_conf, name=f'{self.name}_oq')

    @property
    def enable_iq(self):
        return self._enable_iq

    @property
    def enable_oq(self):
        return self._enable_oq

    @property
    def oq(self):
        if not self.enable_oq:
            raise AttributeError(f'oq has been disabled for {self.name}.')
        return self._oq

    @property
    def beta(self):
        if self._beta is None:
            return ops.cast(0, 'float32')
        return ops.cast(self._beta, ops.dtype(self._beta))

    @property
    def ebops(self):
        if self._ebops is None:
            return ops.cast(0, 'uint32')
        return ops.cast(self._ebops, ops.dtype(self._ebops))

    @property
    def enable_ebops(self):
        return self._enable_ebops

    def build(self, *args, **kwargs):
        super().build(*args, **kwargs)
        if self.enable_ebops:
            self._beta = self.add_weight(
                name='beta',
                shape=(),
                initializer=self._beta0,
                trainable=False,
            )
            self._ebops = self.add_weight(
                name='ebops',
                shape=(),
                initializer=Constant(0.0),
                trainable=False,
                dtype='uint32',
            )
        else:
            self._beta = None
            self._ebops = None

        if not self.enable_oq or self.oq.built:
            return
        self.try_build_output_quantizer(*args, **kwargs)

    def try_build_output_quantizer(self, input_shape=None, *args, **kwargs):
        try:
            output_shape = self.compute_output_shape(input_shape)
        except Exception:
            return
        self.oq.build(output_shape)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                'enable_ebops': self.enable_ebops,
                'beta0': self._beta0,
                'enable_oq': self.enable_oq,
                'enable_iq': self.enable_iq,
                'oq_conf': self.oq.config if self.enable_oq else None,
            }
        )
        return config

    def enable_lora(self, *args, **kwargs):
        raise NotImplementedError('LoRA is not supported in HGQ2.')

    @classmethod
    def from_config(cls, config):
        config = deserialize_keras_object(config)
        return super().from_config(config)

    def _compute_ebops(self, *args, **kwargs):
        raise NotImplementedError('This method is abstract and should be implemented in subclasses.')

    def _post_build(self):
        if self._enable_oq:
            assert hasattr(self, '_oq'), f'Output Quantizer is not defined for {self.name}, but enable_oq is True.'
        if self._enable_iq:
            assert hasattr(self, '_iq'), f'Input Quantizer is not defined for {self.name}, but enable_iq is True.'
        for sublayer in self._flatten_layers():
            assert sublayer.built, f'Sublayer {sublayer.name} is not built for {self.name}'


class QLayerBaseSingleInput(QLayerBase):
    def __init__(
        self,
        iq_conf: QuantizerConfig | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if self.enable_iq:
            iq_conf = iq_conf or QuantizerConfig('default', 'datalane')
            self._iq = Quantizer(iq_conf, name=f'{self.name}_iq')

    @property
    def iq(self):
        if not self.enable_iq:
            raise AttributeError(f'iq has been disabled for {self.name}.')
        return self._iq

    def build(self, input_shape):
        super().build(input_shape)
        if self.enable_iq and not self._iq.built:
            self.iq.build(input_shape)

    def get_config(self):
        config = super().get_config()
        config['iq_conf'] = self.iq.config if self.enable_iq else None
        return config


class MultipleQuantizers(Layer):
    # Not subclassing tuple for stupid tensorflow tracing issue
    def __init__(self, configs: Sequence[QuantizerConfig], **kwargs):
        super().__init__(**kwargs)
        self.quantizers = tuple(Quantizer(config) for config in configs)

    def call(self, x, **kwargs):
        assert len(self) == len(x), f'{self.name} ({self.__class__.__name__}) expects {len(self)} inputs, got {len(x)}.'
        return tuple(f(x_, **kwargs) for f, x_ in zip(self, x))

    def build(self, shapes: Sequence[Sequence[int | None]]):  # type: ignore
        for q, shape in zip(self.quantizers, shapes):
            q.build(shape)
        self.built = True

    def __bool__(self):
        return len(self.quantizers) > 0

    def __len__(self):
        return len(self.quantizers)

    def __getitem__(self, item):
        return self.quantizers[item]

    def __iter__(self):
        return iter(self.quantizers)


class QLayerBaseMultiInputs(QLayerBase):
    def __init__(
        self,
        iq_confs: Sequence[QuantizerConfig] | QuantizerConfig | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._iqs_confs = None
        if self.enable_iq:
            self._iq_confs = iq_confs if iq_confs is not None else QuantizerConfig('default', 'datalane')

    @property
    def iq(self):
        if not self.enable_iq:
            raise AttributeError(f'iq has been disabled for {self.name}.')
        if not self.built:
            raise AttributeError(f'iqs is not available before build for {self.name}.')
        return self._iq

    @property
    def iq_confs(self):
        return self._iq_confs

    def build(self, input_shape):
        super().build(input_shape)
        n_input = len(input_shape)
        for _input_shape in input_shape:
            assert isinstance(_input_shape, Iterable), f'each element of input_shape must be iterable, got {_input_shape}'

        if isinstance(self.iq_confs, QuantizerConfig):
            self._iq_confs = [self.iq_confs] * n_input
        assert len(self.iq_confs) == n_input, (
            f'number of iq_confs must match number of inputs, got {len(self._iq_confs)} != {n_input}'
        )

        self._iq = MultipleQuantizers(self._iq_confs)  # type: ignore
        self._iq.build(input_shape)

    def get_config(self):
        config = super().get_config()
        config['iq_confs'] = self.iq_confs if self.enable_iq else None
        return config



from typing import overload

from keras import ops
from keras.layers import Layer
from keras.saving import deserialize_keras_object

from .config import QuantizerConfig, all_quantizer_keys


class Quantizer(Layer):
    """The generic quantizer layer, wraps internal quantizers to provide a universal interface. Supports float, fixed-point (KBI, KIF) quantization. Can be initialized with a QuantizerConfig object or with the quantizer type and its parameters."""

    @overload
    def __init__(self, config: QuantizerConfig, **kwargs): ...

    @overload
    def __init__(self, q_type='default', place='datalane', **kwargs): ...

    def __init__(self, *args, **kwargs):
        self.supports_masking = True
        self.config, kwargs = self.get_quantizer_config_kwargs(*args, **kwargs)
        self.qnoise_factor = self.config.qnoise_factor
        self.scaler = self.config.scaler
        self.affine = self.config.affine
        super().__init__(**kwargs)
        self.quantizer = self.config.get_quantizer()

    def build(self, input_shape):
        self.quantizer.build(input_shape)
        super().build(input_shape)

    def get_quantizer_config_kwargs(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], QuantizerConfig):
            return args[0], kwargs
        config = kwargs.pop('config', None)
        if isinstance(config, QuantizerConfig):
            return config, kwargs
        else:
            assert config is None, f'config must be a QuantizerConfig object, got {config}'

        _kwargs = {}
        for k in list(kwargs.keys()):
            if k in all_quantizer_keys:
                _kwargs[k] = kwargs.pop(k)
        config = QuantizerConfig(*args, **_kwargs)
        return config, kwargs

    def call(self, inputs, training=None):
        if self.scaler is not None:
            inputs = inputs / self.scaler
        inputs = ops.cast(inputs, ops.dtype(inputs))  # cast to tensor, for sure... (tf is playing naughty here)
        outputs = self.quantizer.call(inputs, training=training)
        if self.scaler is not None:
            outputs = outputs * self.scaler  # type: ignore
        if self.qnoise_factor is not None and training:
            outputs = inputs + self.qnoise_factor * (outputs - inputs)  # type: ignore
        if self.affine is not None:
            outputs = outputs * self.affine[0] + self.affine[1]  # type: ignore
        return outputs

    def get_config(self):
        config = super().get_config()
        config['config'] = self.config
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**deserialize_keras_object(config))

    @property
    def bits(self):
        return self.quantizer.bits

    @property
    def fbits(self):
        return self.quantizer.fbits

    @property
    def q_type(self):
        return self.config.q_type

    def __repr__(self):
        return f'{self.__class__.__name__}(q_type={self.config.q_type}, name={self.name}, built={self.built})'

    def bits_(self, shape):
        bits = self.bits
        return self.quantizer.bw_mapper.bw_to_x(bits, shape)

    def fbits_(self, shape):
        fbits = self.fbits
        return self.quantizer.bw_mapper.bw_to_x(fbits, shape)

    def min_(self, shape):
        _min = self.quantizer.min
        return self.quantizer.bw_mapper.bw_to_x(_min, shape)

    def max_(self, shape):
        _max = self.quantizer.max
        return self.quantizer.bw_mapper.bw_to_x(_max, shape)

    def epsilon_(self, shape):
        epsilon = self.quantizer.epsilon
        return self.quantizer.bw_mapper.bw_to_x(epsilon, shape)



from keras import ops
from keras.layers import EinsumDense

from ...quantizer import Quantizer
from ...quantizer.config import QuantizerConfig
from ...utils.misc import gather_vars_to_kwargs
from .base import QLayerBaseSingleInput


class QEinsumDense(QLayerBaseSingleInput, EinsumDense):
    def __init__(
        self,
        equation,
        output_shape,
        activation=None,
        bias_axes=None,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        parallelization_factor=-1,
        kq_conf: None | QuantizerConfig = None,
        iq_conf: None | QuantizerConfig = None,
        bq_conf: None | QuantizerConfig = None,
        **kwargs,
    ):
        kwargs = gather_vars_to_kwargs('self|kq_conf|bq_conf|parallelization_factor')
        super().__init__(lora_rank=None, **kwargs)

        kq_conf = kq_conf or QuantizerConfig('default', 'weight')
        self._kq = Quantizer(kq_conf, name=f'{self.name}_kq')
        bq_conf = bq_conf or QuantizerConfig('default', 'bias')
        self._bq = None if bias_axes is None else Quantizer(bq_conf, name=f'{self.name}_bq')
        self.parallelization_factor = parallelization_factor

    @property
    def kq(self):
        return self._kq

    @property
    def bq(self):
        return self._bq

    def build(self, input_shape):
        super().build(input_shape)

        self.kq.build(ops.shape(self._kernel))
        if self.bias is not None:
            assert self.bq is not None
            self.bq.build(ops.shape(self.bias))
        self.ebops_equation = self.equation.split('->')[0] + '->'

    def call(self, inputs, training=None):
        qkernel = self.kq(self._kernel, training=training)
        if self.enable_iq:
            inputs = self.iq(inputs, training=training)
        x = ops.einsum(self.equation, inputs, qkernel)
        if self.bias is not None:
            assert self.bq is not None
            x += self.bq(self.bias, training=training)
        if self.activation is not None:
            x = self.activation(x)

        return x

    def _compute_ebops(self, shape):
        # shape = shapes[0]
        bw_inp = self.iq.bits_(shape)
        bw_ker = self.kq.bits_(ops.shape(self.kernel))
        ebops = ops.einsum(self.ebops_equation, bw_inp, bw_ker)
        if self.bq is not None:
            bw_bias = self.bq.bits_(ops.shape(self.bias))
            size = ops.cast(ops.prod(shape), self.dtype)
            ebops = ebops + ops.mean(bw_bias) * size  # type: ignore
        return ebops

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                'kq_conf': self.kq.config,
                'bq_conf': self.bq.config if self.bq is not None else None,
            }
        )
        return config

    @property
    def qkernel(self):
        return self.kq(self._kernel)

    @property
    def qbias(self):
        if self.bias is None:
            return None
        assert self.bq is not None
        return self.bq(self.bias)



from collections.abc import Sequence
from math import prod

from keras import ops
from keras.src import backend

from ..quantizer import QuantizerConfig
from .activation import QUnaryFunctionLUT
from .core import QLayerBaseSingleInput


class QSoftmax(QLayerBaseSingleInput):
    def __init__(
        self,
        axis: int | Sequence[int] | None = None,
        iq_conf: None | QuantizerConfig = None,
        stable=True,
        exp_iq_conf: None | QuantizerConfig = None,
        exp_oq_conf: None | QuantizerConfig = None,
        inv_iq_conf: None | QuantizerConfig = None,
        inv_oq_conf: None | QuantizerConfig = None,
        allow_heterogeneous_table: bool = False,
        input_scaler: float = 1.0,
        parallelization_factor: int = -1,
        **kwargs,
    ):
        # Keras h5 loader pops axis silent when it is a list longer than 1.
        axes = kwargs.pop('axes', None)
        assert axis is None or axes is None, 'Use only one of `axis` or `axes`.'
        self.axes = axes or (tuple(axis) if isinstance(axis, Sequence) else (axis,) if axis is not None else (-1,))

        self.supports_masking = True
        super().__init__(iq_conf=iq_conf, **kwargs)  # type: ignore
        self.stable = stable
        self.parallelization_factor = parallelization_factor

        self._allow_heterogeneous_table = allow_heterogeneous_table

        self.input_scaler = input_scaler

        def _inv(x):
            return 1.0 / (x + backend.epsilon())

        def _exp(x):
            if self.stable:
                return ops.exp(-x * self.input_scaler)
            else:
                return ops.exp(x * self.input_scaler)

        inv_iq_conf = inv_iq_conf or QuantizerConfig('default', 'datalane')
        exp_iq_conf = exp_iq_conf or QuantizerConfig('default', 'datalane')
        exp_oq_conf = exp_oq_conf or QuantizerConfig('default', 'table')
        inv_oq_conf = inv_oq_conf or QuantizerConfig('default', 'table')
        if not self._allow_heterogeneous_table:
            inv_iq_conf.config['heterogeneous_axis'] = ()
            inv_iq_conf.config['homogeneous_axis'] = None
            exp_iq_conf.config['heterogeneous_axis'] = ()
            exp_iq_conf.config['homogeneous_axis'] = None

        if 'k0' in inv_oq_conf.config:
            inv_oq_conf.config['k0'] = 0
        if 'k0' in exp_oq_conf.config:
            exp_oq_conf.config['k0'] = 0

        # hls4ml only supports the following configurations
        inv_oq_conf.config['overflow_mode'] = 'SAT'  # type: ignore
        inv_oq_conf.config['round_mode'] = 'RND_CONV'  # type: ignore
        exp_oq_conf.config['overflow_mode'] = 'SAT'  # type: ignore
        exp_oq_conf.config['round_mode'] = 'RND_CONV'  # type: ignore

        self.inv_table = QUnaryFunctionLUT(
            _inv,
            inv_iq_conf,
            inv_oq_conf,
            enable_iq=True,
            enable_oq=True,
            allow_heterogeneous_table=allow_heterogeneous_table,
            name=f'{self.name}_inv_table',
            enable_ebops=self.enable_ebops,
            beta0=self._beta0.clone(),
        )
        self.exp_table = QUnaryFunctionLUT(
            _exp,
            exp_iq_conf,
            exp_oq_conf,
            enable_iq=self.stable,
            enable_oq=True,
            allow_heterogeneous_table=allow_heterogeneous_table,
            allow_heterogeneous_input=True,
            name=f'{self.name}_exp_table',
            enable_ebops=self.enable_ebops and stable,
            beta0=self._beta0.clone(),
        )

    def build(self, input_shape):
        self.exp_table.build(input_shape)
        axis = sorted(i if i >= 0 else i + len(input_shape) for i in self.axes)  # type: ignore
        self.axes = tuple(axis)

        inv_shape = list(input_shape)
        for i in self.axes:
            inv_shape[i] = 1
        self.inv_table.build(tuple(inv_shape))

        super().build(input_shape)

    def call(self, inputs, training=None, mask=None):  # type: ignore
        if self.enable_iq:
            inputs = self.iq(inputs, training=training)

        if self.stable:
            inputs = ops.max(inputs, axis=self.axes, keepdims=True) - inputs

        exp_inp = self.exp_table(inputs, training=training)

        if mask is not None:
            exp_inp = backend.cast(mask, ops.dtype(inputs)) * exp_inp

        sums = ops.sum(exp_inp, axis=self.axes, keepdims=True)  # type: ignore
        divisor = self.inv_table(sums, training=training)

        return exp_inp * divisor

    def _compute_ebops(self, shape):
        accum_shape = tuple(1 if i in self.axes else s for i, s in enumerate(shape))
        max_instance = prod(accum_shape)
        n_instance = self.parallelization_factor if self.parallelization_factor > 0 else max_instance
        factor = n_instance / max_instance

        inp_bits = self.iq.bits_(shape)
        exp_bits = self.exp_table.oq.bits_(shape)
        inv_bits = self.inv_table.oq.bits_(accum_shape)

        if self.stable:
            substract_ebops = ops.sum(inp_bits)  # type: ignore # TODO: better ebops cost model for add and max
        else:
            substract_ebops = 0

        accum_ebops = ops.sum(exp_bits) - ops.sum(ops.min(exp_bits, axis=self.axes))  # type: ignore
        mult_ebops = ops.sum(exp_bits * inv_bits)  # type: ignore

        ebops = substract_ebops + accum_ebops + mult_ebops

        if not self.stable:
            # iq is disabled for exp table, compute here
            ebops += ops.sum((2.0**inp_bits) * exp_bits) * 1e-4  # type: ignore

        return ebops * factor

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                'axes': self.axes,
                'stable': self.stable,
                'exp_oq_conf': self.exp_table.oq.config,
                'exp_iq_conf': self.exp_table.iq.config if self.stable else None,
                'inv_oq_conf': self.inv_table.oq.config,
                'inv_iq_conf': self.inv_table.iq.config,
                'allow_heterogeneous_table': self._allow_heterogeneous_table,
                'input_scaler': self.input_scaler,
                'parallelization_factor': self.parallelization_factor,
            }
        )
        return config

    def compute_output_shape(self, input_shape):
        return input_shape

    @property
    def ebops(self):
        ebops = sum(
            (  # type: ignore
                ops.convert_to_tensor(self._ebops),
                self.exp_table.ebops,
                self.inv_table.ebops,
            )
        )
        return round(ops.convert_to_numpy(ebops).item())  # type: ignore



from .quantized_bits import quantized_bits, quantized_relu

__all__ = ['quantized_bits', 'quantized_relu', 'get_quantizer']


def get_quantizer(str_conf: str):
    name = str_conf.split('(', 1)[0]
    if name in globals():
        return globals()[name].from_string(str_conf)
    raise ValueError(f'Unknown quantizer: {name}')



from keras import ops
from keras.initializers import Initializer

from hgq.quantizer import Quantizer


class QInitializer(Initializer):
    """QKeras compatible initializer."""

    def __init__(self, initializer: Initializer, quantizer: Quantizer):
        self.initializer = initializer
        self.quantizer = quantizer

    def __call__(self, shape, dtype=None):
        x = self.initializer(shape, dtype)

        # max_x = ops.max(ops.abs(x))
        std_x = ops.std(x)
        delta = ops.mean(self.quantizer.epsilon_(ops.shape(x)))

        # delta is the minimum resolution of the number system.
        # we want to make sure we have enough values.
        if delta > std_x and hasattr(self.initializer, 'scale'):  # type: ignore
            q = self.quantizer(x)
            max_q = ops.max(ops.abs(q))
            scale = 1.0
            if max_q == 0.0:
                xx = ops.mean(x * x)
                scale = ops.mean(self.quantizer.max_(ops.shape(x))) / ops.sqrt(xx)  # type: ignore
            else:
                qx = ops.sum(q * x)
                qq = ops.sum(q * q)

                scale = qq / qx  # type: ignore

            self.initializer.scale *= max(scale, 1)  # type: ignore
            x = self.initializer(shape, dtype)

        _min, _max = self.quantizer.min_(ops.shape(x)), self.quantizer.max_(ops.shape(x))
        _min = ops.max(_min, -_max)  # type: ignore
        return ops.clip(x, _min, _max)

    def get_config(self):
        # TODO: either avoid (stateful) quantizers in the class, or make the init ephemeral. Avoid saving (stateful) quantizers.
        # Error is tentatively raised to avoid saving (stateful) quantizers.
        raise ValueError('QInitializer is ephemeral in hgq/qkeras. It should not be saved.')



from ..quantizer.config import (
    HardSigmoidConfig,
    HardTanhConfig,
    QuantizerConfig,
    QuantizerConfigScope,
    all_quantizer_types,
    default_configs,
)

__all__ = [
    'QuantizerConfig',
    'QuantizerConfigScope',
    'all_quantizer_types',
    'default_configs',
    'HardSigmoidConfig',
    'HardTanhConfig',
]



from math import log2

from keras import ops
from keras.layers import Add, Average, Dot, Maximum, Minimum, Multiply, Subtract
from keras.src.layers.merging.base_merge import Merge

from ..core.base import QLayerBaseMultiInputs


class QMerge(QLayerBaseMultiInputs, Merge):  # type: ignore
    def call(self, inputs, training=None):
        if self.enable_iq:
            inputs = self.iq(inputs, training=training)
        r = super().call(inputs)
        return r


def _ebops_from_sum_bits_excl_max(self: QMerge, shapes):
    bitss = [iq.bits_(shape) for iq, shape in zip(self.iq, shapes)]
    _ebops = ops.sum(sum(bitss))
    _min = bitss[0]
    for bits in bitss[1:]:
        _min = ops.minimum(_min, bits)
    ebops = _ebops - ops.sum(_min)  # type: ignore
    return ebops


class QAdd(QMerge, Add):
    def _compute_ebops(self, *shapes):
        return _ebops_from_sum_bits_excl_max(self, shapes) * 0.65


class QAveragePow2(QAdd, Average):
    def build(self, input_shape):
        super().build(input_shape)
        self._scale = 2.0 ** round(log2(1.0 / len(input_shape)))

    def _merge_function(self, inputs):
        output = inputs[0]
        for i in range(1, len(inputs)):
            output = ops.add(output, inputs[i])
        return output * self._scale  # type: ignore


class QSubtract(QMerge, Subtract):
    def _compute_ebops(self, *shapes):
        bits0, bits1 = (iq.bits_(shape) for iq, shape in zip(self.iq, shapes))
        ebops = ops.sum(ops.maximum(bits0, bits1))
        ebops = ebops  # TODO: better ebops cost model for subtract
        return ebops


class QMultiply(QMerge, Multiply):
    def _compute_ebops(self, *shapes):
        bitss = [iq.bits_(shape) for iq, shape in zip(self.iq, shapes)]
        _ebops = bitss[0]
        for bits in bitss[1:]:
            _ebops = ops.multiply(_ebops, bits)
        ebops = ops.sum(_ebops)
        return ebops


class QDot(QMerge, Dot):
    def __init__(self, axes=-1, **kwargs):
        super().__init__(axes=axes, **kwargs)
        assert not self.normalize, 'normalize is not supported'

    def build(self, input_shape):
        super().build(input_shape)

    def _compute_ebops(self, *shapes):
        bits0, bits1 = (iq.bits_(shape) for iq, shape in zip(self.iq, shapes))
        ebops = ops.sum(bits0 * bits1)
        return ebops


class QMaximum(QMerge, Maximum):
    def _compute_ebops(self, *shapes):
        return 1.0 * _ebops_from_sum_bits_excl_max(self, shapes)


class QMinimum(QMerge, Minimum):
    def _compute_ebops(self, *shapes):
        return 1.0 * _ebops_from_sum_bits_excl_max(self, shapes)



from collections.abc import Callable

import keras
from keras import ops
from keras.layers import Layer
from keras.saving import register_keras_serializable

from ...quantizer import Quantizer, QuantizerConfig
from ..core import QLayerBaseSingleInput


@register_keras_serializable(package='hgq2')
class QDenseT(QLayerBaseSingleInput):
    def __init__(
        self,
        n_out: int,
        n_hl: int,
        d_hl: int,
        subnn_activation: str | Callable | Layer | None,
        activation: Callable | None | str = None,
        toq_conf: QuantizerConfig | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.n_out = n_out
        self.d_hl = d_hl
        self.subnn_activation = keras.activations.get(subnn_activation)

        assert n_hl >= 0

        self.n_hl = n_hl

        toq_conf = toq_conf or QuantizerConfig(place='table')
        self._toq = Quantizer(toq_conf)
        self.activation = keras.activations.get(activation)

    def _build_module(self, n_in: int):
        layers = []
        _shape = (n_in, self.n_out, self.d_hl)
        for _ in range(self.n_hl):
            layers.append(keras.layers.EinsumDense('biod,iodD->bioD', _shape, self.subnn_activation, bias_axes='ioD'))
        l_out = keras.layers.EinsumDense('biod,iod->bio', (n_in, self.n_out), 'linear', bias_axes='io')
        layers.append(l_out)
        module = keras.models.Sequential(layers)
        return module

    def call(self, x, training=None):
        n_in = keras.ops.shape(x)[-1]
        x = keras.ops.broadcast_to(x[..., None], (keras.ops.shape(x)[0], n_in, self.n_out))  # (B, N_in, N_out)
        if self.enable_iq:
            x = self.iq(x)
        x = x[..., None]  # (B, N_in, N_out, 1)

        return self.activation(ops.sum(self.toq(self.module(x, training=training)), axis=1))

    def build(self, input_shape):
        input_shape = tuple(input_shape)
        self.n_in = input_shape[-1]
        if self.enable_iq and not self.iq.built:
            self.iq.build(input_shape + (self.n_out,))
        self.toq.build(input_shape + (self.n_out,))
        self.module = self._build_module(self.n_in)
        self.module.build(input_shape + (self.n_out, 1))
        super().build(input_shape)

    def _compute_ebops(self, shape: tuple[int, ...]):
        q_shape = shape + (self.n_out,)
        bits_in = self.iq.fbits_(q_shape)
        bits_out = self.toq.fbits_(q_shape)

        eff_lut5_count = ops.where(bits_in >= 5, 2 ** (bits_in - 5), 0.2 * bits_in)  # type: ignore

        table_lut5s = ops.dot(ops.ravel(eff_lut5_count), ops.ravel(bits_out)) * 0.5  # type: ignore

        return table_lut5s + ops.sum(bits_out)

    @property
    def toq(self):
        return self._toq

    def get_config(self):
        config = {
            'n_out': self.n_out,
            'n_hl': self.n_hl,
            'd_hl': self.d_hl,
            'subnn_activation': self.subnn_activation,
            'activation': self.activation,
            'toq_conf': self.toq.config,
            **super().get_config(),
        }
        return config



import keras
import numpy as np
import pytest
from keras import ops

from hgq.layers import QLinformerAttention, QMultiHeadAttention
from hgq.quantizer.internal import FixedPointQuantizerKBI, FixedPointQuantizerKIF
from tests.base import LayerTestBase


class TestMultiHeadAttention(LayerTestBase):
    layer_cls = QMultiHeadAttention

    @pytest.fixture(params=[1, 2])  # Test different head counts
    def num_heads(self, request):
        return request.param

    @pytest.fixture(params=[8])  # Test different key dimensions
    def key_dim(self, request):
        return request.param

    @pytest.fixture(params=['none'])  # Test different fusion modes
    def fuse(self, request):
        # qkv and kv fused projection not implemented in hls4ml yet
        return request.param

    @pytest.fixture
    def input_shapes(self, fuse: str):
        # Seq..., dim
        if fuse == 'none':
            return (2, 4, 8), (2, 4, 9), (2, 4, 10)
        elif fuse == 'qkv':
            return (2, 4, 8)
        elif fuse == 'kv':
            return (2, 4, 3), (2, 4, 5)
        raise ValueError(f'Invalid fusion mode: {fuse}')

    @pytest.fixture
    def value_dim(self, key_dim):
        return key_dim if self.fuse in ('qkv', 'kv') else key_dim + 1

    @pytest.fixture
    def layer_kwargs(self, num_heads, key_dim, fuse):
        return {'num_heads': num_heads, 'key_dim': key_dim, 'fuse': fuse}

    @pytest.fixture(params=[True])
    def use_parallel_io(self, request) -> bool:
        return request.param

    @pytest.fixture
    def model(self, layer, input_shapes, use_parallel_io):
        """Create test model with the layer"""
        if isinstance(input_shapes[0], int):
            input_shapes = (input_shapes,)
        inputs = [keras.layers.Input(shape=shape) for shape in input_shapes]

        outputs = layer(*inputs)  # Differs to other layers, MHA takes 2-3 inputs (q, v, k=v) not in a list
        model = keras.Model(inputs, outputs)

        self.perturbe_bw(use_parallel_io, model)
        return model

    def perturbe_bw(self, use_parallel_io, model):
        if use_parallel_io:
            for _layer in model._flatten_layers(False):
                if isinstance(_layer, FixedPointQuantizerKBI):
                    b = np.random.randint(4, 8, _layer._b.shape)
                    i = ops.convert_to_numpy(ops.stop_gradient(_layer.i))
                    b = np.minimum(b, 12 - i)
                    if np.all(b == 0):
                        b.ravel()[0] = 1
                    _layer._b.assign(ops.array(b))
                if isinstance(_layer, FixedPointQuantizerKIF):
                    f = np.random.randint(2, 8, _layer._f.shape)
                    i = ops.convert_to_numpy(ops.stop_gradient(_layer.i))
                    f = np.minimum(f, 12 - i)
                    if np.all(i + f == 0):
                        f.ravel()[0] = 1
                    _layer._f.assign(ops.array(f))
        for _layer in model._flatten_layers(False):
            # Randomize activation values
            if hasattr(_layer, 'bias') and isinstance(_layer.bias, keras.Variable):
                bias = np.random.randn(*_layer.bias.shape)
                _layer.bias.assign(ops.array(bias))

    @pytest.fixture
    def input_data(self, input_shapes, N: int = 5000):
        return tuple(np.random.randn(N, *shape).astype(np.float32) * 3 for shape in input_shapes)

    def assert_equal(self, keras_output, hls_output):
        return np.testing.assert_allclose(keras_output, hls_output, atol=1e-6)


class TestLinformerAttention(TestMultiHeadAttention):
    layer_cls = QLinformerAttention

    @pytest.fixture
    def layer_kwargs(self, num_heads, key_dim, fuse, input_shapes):
        lin_kv_proj_dim = [2 for _ in range(len(input_shapes[0]) - 1)]
        return {'num_heads': num_heads, 'key_dim': key_dim, 'fuse': fuse, 'lin_kv_proj_dim': lin_kv_proj_dim}



from keras import ops
from keras.layers import GRU, GRUCell
from keras.saving import register_keras_serializable
from keras.src import tree
from keras.src.layers.input_spec import InputSpec

from ...config import HardSigmoidConfig, HardTanhConfig, QuantizerConfig
from ...quantizer import Quantizer
from ..core.base import QLayerBase
from .simple_rnn import QRNN


@register_keras_serializable(package='hgq')
class QGRUCell(QLayerBase, GRUCell):
    """Cell class for the GRU layer.

    This class processes one step within the whole time sequence input, whereas
    `keras.layer.GRU` processes the whole sequence.

    Parameters
    ----------
    units : int
        Positive integer, dimensionality of the output space.
    activation : str, optional
        Activation function to use. Default: hyperbolic tangent
        (`tanh`). If you pass None, no activation is applied
        (ie. "linear" activation: `a(x) = x`).
    recurrent_activation : str, optional
        Activation function to use for the recurrent step.
        Default: sigmoid (`sigmoid`). If you pass `None`, no activation is
        applied (ie. "linear" activation: `a(x) = x`).
    use_bias : bool, optional
        Whether the layer should use a bias vector. Default: True.
    kernel_initializer : str, optional
        Initializer for the `kernel` weights matrix,
        used for the linear transformation of the inputs. Default: "glorot_uniform".
    recurrent_initializer : str, optional
        Initializer for the `recurrent_kernel` weights matrix, used for the linear transformation
        of the recurrent state. Default: "orthogonal".
    bias_initializer : str, optional
        Initializer for the bias vector. Default: "zeros".
    kernel_regularizer : optional
        Regularizer function applied to the `kernel` weights matrix. Default: None.
    recurrent_regularizer : optional
        Regularizer function applied to the `recurrent_kernel` weights matrix. Default: None.
    bias_regularizer : optional
        Regularizer function applied to the bias vector. Default: None.
    kernel_constraint : optional
        Constraint function applied to the `kernel` weights matrix. Default: None.
    recurrent_constraint : optional
        Constraint function applied to the `recurrent_kernel` weights matrix. Default: None.
    bias_constraint : optional
        Constraint function applied to the bias vector. Default: None.
    dropout : float, optional
        Float between 0 and 1. Fraction of the units to drop for the
        linear transformation of the inputs. Default: 0.
    recurrent_dropout : float, optional
        Float between 0 and 1. Fraction of the units to drop for the
        linear transformation of the recurrent state. Default: 0.
    reset_after : bool, optional
        GRU convention (whether to apply reset gate after or
        before matrix multiplication). False = "before",
        True = "after" (default and cuDNN compatible).
    seed : int, optional
        Random seed for dropout.
    iq_conf : QuantizerConfig or None, optional
        Quantizer configuration for input quantizer. Default: None.
    paq_conf : QuantizerConfig or None, optional
        Quantizer configuration for post-activation quantizer. Default: None.
    praq_conf : QuantizerConfig or None, optional
        Quantizer configuration for pre-recurrent activation quantizer. Default: None.
    sq_conf : QuantizerConfig or None, optional
        Quantizer configuration for state quantizer. Default: None.
    kq_conf : QuantizerConfig or None, optional
        Quantizer configuration for kernel quantizer. Default: None.
    rkq_conf : QuantizerConfig or None, optional
        Quantizer configuration for recurrent kernel quantizer. Default: None.
    bq_conf : QuantizerConfig or None, optional
        Quantizer configuration for bias quantizer. Default: None.
    oq_conf : QuantizerConfig or None, optional
        Quantizer configuration for output quantizer. Default: None.
    rhq_conf : QuantizerConfig or None, optional
        Quantizer configuration for recurrent hidden state quantizer. Default: None.
    standalone : bool, optional
        Whether this cell is used standalone or as part of a larger RNN layer.
        EBOPS computation will be skipped when used as a sublayer.
        Default: True.
    enable_ebops : bool or None, optional
        Whether to enable energy-efficient bit operations. Default: None.
    enable_iq : bool or None, optional
        Whether to enable input quantizer. Default: None.
    enable_oq : bool or None, optional
        Whether to enable output quantizer. Default: None.

    Notes
    -----
    inputs : array_like
        A 2D tensor, with shape `(batch, features)`.
    states : array_like
        A 2D tensor with shape `(batch, units)`, which is the state
        from the previous time step.
    training : bool, optional
        Python boolean indicating whether the layer should behave in
        training mode or in inference mode. Only relevant when `dropout` or
        `recurrent_dropout` is used.
    """

    def __init__(
        self,
        units,
        activation='linear',
        recurrent_activation='linear',
        use_bias=True,
        kernel_initializer='glorot_uniform',
        recurrent_initializer='orthogonal',
        bias_initializer='zeros',
        kernel_regularizer=None,
        recurrent_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        recurrent_constraint=None,
        bias_constraint=None,
        dropout=0.0,
        recurrent_dropout=0.0,
        reset_after=True,
        seed=None,
        iq_conf: QuantizerConfig | None = None,
        paq_conf: QuantizerConfig | None = None,
        praq_conf: QuantizerConfig | None = None,
        sq_conf: QuantizerConfig | None = None,
        kq_conf: QuantizerConfig | None = None,
        rkq_conf: QuantizerConfig | None = None,
        bq_conf: QuantizerConfig | None = None,
        oq_conf: QuantizerConfig | None = None,
        rhq_conf: QuantizerConfig | None = None,
        standalone: bool = True,
        enable_ebops: bool | None = None,
        enable_iq: bool | None = None,
        enable_oq: bool | None = None,
        **kwargs,
    ):
        super().__init__(
            units=units,
            activation=activation,
            recurrent_activation=recurrent_activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint,
            bias_constraint=bias_constraint,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            reset_after=reset_after,
            seed=seed,
            oq_conf=oq_conf,
            enable_oq=enable_oq,
            enable_iq=enable_iq,
            enable_ebops=enable_ebops,
            **kwargs,
        )

        paq_conf = paq_conf or HardTanhConfig(place='datalane')
        praq_conf = praq_conf or HardSigmoidConfig(place='datalane')
        iq_conf = iq_conf or QuantizerConfig(place='datalane')
        sq_conf = sq_conf or QuantizerConfig(place='datalane')
        kq_conf = kq_conf or QuantizerConfig(place='weight')
        rkq_conf = rkq_conf or QuantizerConfig(place='weight')
        bq_conf = bq_conf or QuantizerConfig(place='bias')
        rhq_conf = rhq_conf or QuantizerConfig(place='datalane')
        self.standalone = standalone

        if self._enable_iq:
            self._iq = Quantizer(iq_conf, name=f'{self.name}_iq')
        self._paq = Quantizer(paq_conf, name=f'{self.name}_paq')
        self._praq = Quantizer(praq_conf, name=f'{self.name}_praq')
        self._sq = Quantizer(sq_conf, name=f'{self.name}_sq')
        self._kq = Quantizer(kq_conf, name=f'{self.name}_kq')
        self._rkq = Quantizer(rkq_conf, name=f'{self.name}_rkq')
        if self.use_bias:
            self._bq = Quantizer(bq_conf, name=f'{self.name}_bq')
        self._rhq = Quantizer(rhq_conf, name=f'{self.name}_rhq')

    @property
    def paq(self):
        return self._paq

    @property
    def praq(self):
        return self._praq

    @property
    def kq(self):
        return self._kq

    @property
    def bq(self):
        return self._bq

    @property
    def rkq(self):
        return self._rkq

    @property
    def iq(self):
        if not self.enable_iq:
            raise AttributeError(f'iq has been disabled for {self.name}.')
        return self._iq

    @property
    def sq(self):
        return self._sq

    @property
    def rhq(self):
        return self._rhq

    @property
    def qkernel(self):
        return self.kq(self.kernel)

    @property
    def qrecurrent_kernel(self):
        return self.rkq(self.recurrent_kernel)

    @property
    def qbias(self):
        if not self.use_bias:
            raise AttributeError(f'bias has been disabled for {self.name}.')
        assert self.bq is not None
        return self.bq(self.bias)

    def qactivation(self, x):
        return self.paq(self.activation(x))

    def qrecurrent_activation(self, x):
        return self.praq(self.recurrent_activation(x))

    def build(self, input_shape):
        state_shape = (input_shape[0], self.units)
        super().build(input_shape)
        if self._enable_iq:
            self.iq.build(input_shape)
            self.sq.build(state_shape)
        self.kq.build(self.kernel.shape)
        if self.use_bias:
            assert self.bq is not None
            self.bq.build(self.bias.shape)  # type: ignore
        self.rkq.build(self.recurrent_kernel.shape)
        if self._enable_oq:
            self.oq.build(state_shape)
        self.rhq.build(state_shape)
        self.paq.build(state_shape)
        self.praq.build((state_shape[0], 2 * self.units))

    def call(self, inputs, states, training=False):
        h_tm1 = states[0] if tree.is_nested(states) else states  # previous state

        qh_tm1 = self.sq(h_tm1)
        qinputs = self.iq(inputs) if self.enable_iq else inputs

        if self.use_bias:
            if not self.reset_after:
                input_qbias, recurrent_qbias = self.qbias, None
            else:
                input_qbias, recurrent_qbias = self.qbias
        else:
            input_qbias, recurrent_qbias = 0, 0

        if training and 0.0 < self.dropout < 1.0:
            dp_mask = self.get_dropout_mask(qinputs)
            qinputs = qinputs * dp_mask

        matrix_x = qinputs @ self.qkernel + input_qbias

        x_zr = matrix_x[:, : 2 * self.units]
        x_h = matrix_x[:, 2 * self.units :]

        qrecurrent_kernel = self.qrecurrent_kernel
        if self.reset_after:
            # hidden state projected by all gate matrices at once
            matrix_inner = qh_tm1 @ qrecurrent_kernel
            if self.use_bias:
                matrix_inner += recurrent_qbias
        else:
            # hidden state projected separately for update/reset and new
            matrix_inner = qh_tm1 @ qrecurrent_kernel[:, : 2 * self.units]

        recurrent_zr = matrix_inner[:, : 2 * self.units]

        qzr = self.qrecurrent_activation(x_zr + recurrent_zr)
        qz, qr = ops.split(qzr, 2, axis=-1)

        if self.reset_after:
            recurrent_h = qr * self.rhq(matrix_inner[:, self.units * 2 :])
        else:
            recurrent_h = ops.matmul(self.rhq(qr * qh_tm1), qrecurrent_kernel[:, 2 * self.units :])

        qhh = self.qactivation(x_h + recurrent_h)

        # previous and candidate state mixed by update gate
        h = qz * qh_tm1 + (1 - qz) * qhh  # type: ignore
        new_state = [h] if tree.is_nested(states) else h
        return h, new_state

    def get_initial_state(self, batch_size=None):
        return [ops.zeros((batch_size, self.state_size), dtype=self.compute_dtype)]

    def get_config(self):
        config = {
            'paq_conf': self.paq.config,
            'praq_conf': self.praq.config,
            'iq_conf': self.iq.config if self.enable_iq else None,
            'sq_conf': self.sq.config,
            'kq_conf': self.kq.config,
            'rkq_conf': self.rkq.config,
            'bq_conf': self.bq.config if self.use_bias else None,  # type: ignore
            'beta0': self._beta0,
            **super().get_config(),
        }
        return config

    def _compute_ebops(self, shape, state_shape):
        bw_state = self.sq.bits_(state_shape)
        bw_inp = self.iq.bits_(shape) if self.enable_iq else 0
        bw_ker = self.kq.bits_(ops.shape(self.kernel))
        bw_rker = self.rkq.bits_(ops.shape(self.recurrent_kernel))
        bw_rh = self.rhq.bits_(state_shape)
        bw_zr = self.praq.bits_((*shape[:-1], 2 * self.units))

        if self.use_bias:
            bw_bias = self.bq.bits_(ops.shape(self.bias))  # type: ignore
            if not self.reset_after:
                ebops_bias = ops.sum(bw_bias[0])
            else:
                ebops_bias = ops.sum(bw_bias)
        else:
            ebops_bias = 0

        ebops_0 = ops.sum(ops.matmul(bw_inp, bw_ker))

        if self.reset_after:
            ebops1 = ops.sum(ops.matmul(bw_state, bw_rker))
        else:
            ebops1 = ops.sum(ops.matmul(bw_state, bw_rker[:, : 2 * self.units]))

        bw_z, bw_r = ops.split(bw_zr, 2, axis=-1)
        if self.reset_after:
            ebops2 = ops.sum(bw_r * bw_rh)  # type: ignore
        else:
            ebops2 = ops.sum(ops.matmul(bw_rh, bw_rker[:, 2 * self.units :])) + ops.sum(bw_r * bw_state)  # type: ignore

        bw_qhh = self.paq.bits_(state_shape)

        ebops3 = ops.sum(bw_z * (bw_qhh + bw_state))  # type: ignore
        return ebops_0 + ebops1 + ebops2 + ebops3 + ebops_bias  # type: ignore

    @property
    def enable_ebops(self):
        # When used as a sublayer in the RNN layer, standalone is set to False
        # EBOPs computation handled on the higher level RNN layer
        return self._enable_ebops and self.standalone


class QGRU(QRNN, GRU):
    """Gated Recurrent Unit - Cho et al. 2014.

    The QGRU only allows the backend native implementation (no CuDNN kernel).
    When the jax backend is used, if any `WRAP` quantizers are used, unroll will
    be set to `True` to avoid the side effect issue in the `jax.lax.scan` loop.

    Parameters
    ----------
    units : int
        Positive integer, dimensionality of the output space.
    activation : str, optional
        Activation function to use.
        Default: linear, effectively hard_tanh by the pre-activation quantizer.
    recurrent_activation : str, optional
        Activation function to use for the recurrent step.
        Default: linear, effectively hard_sigmoid (slope=0.5) by the pre-activation quantizer.
    use_bias : bool, optional
        Whether the layer should use a bias vector. Default: True.
    kernel_initializer : str, optional
        Initializer for the `kernel` weights matrix,
        used for the linear transformation of the inputs. Default: "glorot_uniform".
    recurrent_initializer : str, optional
        Initializer for the `recurrent_kernel` weights matrix, used for the linear transformation
        of the recurrent state. Default: "orthogonal".
    bias_initializer : str, optional
        Initializer for the bias vector. Default: "zeros".
    kernel_regularizer : optional
        Regularizer function applied to the `kernel` weights matrix. Default: None.
    recurrent_regularizer : optional
        Regularizer function applied to the `recurrent_kernel` weights matrix. Default: None.
    bias_regularizer : optional
        Regularizer function applied to the bias vector. Default: None.
    activity_regularizer : optional
        Regularizer function applied to the output of the layer (its "activation"). Default: None.
    kernel_constraint : optional
        Constraint function applied to the `kernel` weights matrix. Default: None.
    recurrent_constraint : optional
        Constraint function applied to the `recurrent_kernel` weights matrix. Default: None.
    bias_constraint : optional
        Constraint function applied to the bias vector. Default: None.
    dropout : float, optional
        Float between 0 and 1. Fraction of the units to drop for the
        linear transformation of the inputs. Default: 0.
    recurrent_dropout : float, optional
        Float between 0 and 1. Fraction of the units to drop for the
        linear transformation of the recurrent state. Default: 0.
    seed : int, optional
        Random seed for dropout.
    return_sequences : bool, optional
        Whether to return the last output in the output sequence, or the full sequence. Default: False.
    return_state : bool, optional
        Whether to return the last state in addition to the output. Default: False.
    go_backwards : bool, optional
        If True, process the input sequence backwards and return the reversed sequence. Default: False.
    stateful : bool, optional
        If True, the last state for each sample at index i in a batch will be used as initial
        state for the sample of index i in the following batch. Default: False.
    unroll : bool or None, optional
        None is equivalent to False. However, for the JAX backend, if
        any `WRAP` quantizers are used, unroll will be set to True
        to avoid the side effect issue in the `jax.lax.scan` loop.
        If True, the network will be unrolled,
        else a symbolic loop will be used.
        Unrolling can speed-up a RNN,
        although it tends to be more memory-intensive.
        Unrolling is only suitable for short sequences. Default: None.
    reset_after : bool, optional
        GRU convention (whether to apply reset gate after or
        before matrix multiplication). False is "before",
        True is "after" (default and cuDNN compatible). Default: True.
    iq_conf : QuantizerConfig or None, optional
        Quantizer configuration for input quantizer. Default: None (global default)
    paq_conf : QuantizerConfig or None, optional
        Quantizer configuration for post-activation quantizer.
        Default: None (hard tanh like, w/ global default)
    praq_conf : QuantizerConfig or None, optional
        Quantizer configuration for pre-recurrent activation quantizer.
        Default: None (hard sigmoid like, w/ global default)
    sq_conf : QuantizerConfig or None, optional
        Quantizer configuration for state quantizer. Default: None (global default)
    kq_conf : QuantizerConfig or None, optional
        Quantizer configuration for kernel quantizer. Default: None (global default)
    rkq_conf : QuantizerConfig or None, optional
        Quantizer configuration for recurrent kernel quantizer. Default: None (global default)
    bq_conf : QuantizerConfig or None, optional
        Quantizer configuration for bias quantizer. Default: None (global default)
    oq_conf : QuantizerConfig or None, optional
        Quantizer configuration for output quantizer. Default: None (global default)
    rhq_conf : QuantizerConfig or None, optional
        Quantizer configuration for recurrent hidden state quantizer. Default: None (global default)
    parallelization_factor : int, optional
        Factor for parallelization. Default: 1.
    enable_oq : bool or None, optional
        Whether to enable output quantizer. Default: None (global default)
    enable_iq : bool or None, optional
        Whether to enable input quantizer. Default: None (global default)
    enable_ebops : bool or None, optional
        Whether to enable energy-efficient bit operations. Default: None (global default)
    beta0 : float or None, optional
        Beta0 parameter for quantizer. Default: None (global default)
    enable_ebops : bool or None, optional
        Whether to enable EBOPs resource consumption estimation. Default: None (global default).
    parallelization_factor : int, optional
        Number of cells to be computed in parallel. Default: 1.

    Notes
    -----
    inputs : array_like
        A 3D tensor, with shape `(batch, timesteps, feature)`.
    mask : array_like, optional
        Binary tensor of shape `(samples, timesteps)` indicating whether
        a given timestep should be masked (optional).
        An individual `True` entry indicates that the corresponding timestep
        should be utilized, while a `False` entry indicates that the
        corresponding timestep should be ignored. Defaults to `None`.
    training : bool, optional
        Python boolean indicating whether the layer should behave in
        training mode or in inference mode. This argument is passed to the
        cell when calling it. This is only relevant if `dropout` or
        `recurrent_dropout` is used (optional). Defaults to `None`.
    initial_state : list, optional
        List of initial state tensors to be passed to the first
        call of the cell (optional, `None` causes creation
        of zero-filled initial state tensors). Defaults to `None`.
    """

    def __init__(
        self,
        units,
        activation='linear',
        recurrent_activation='linear',
        use_bias=True,
        kernel_initializer='glorot_uniform',
        recurrent_initializer='orthogonal',
        bias_initializer='zeros',
        kernel_regularizer=None,
        recurrent_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        recurrent_constraint=None,
        bias_constraint=None,
        dropout=0.0,
        recurrent_dropout=0.0,
        seed=None,
        return_sequences=False,
        return_state=False,
        go_backwards=False,
        stateful=False,
        unroll=None,
        reset_after=True,
        iq_conf: QuantizerConfig | None = None,
        paq_conf: QuantizerConfig | None = None,
        praq_conf: QuantizerConfig | None = None,
        sq_conf: QuantizerConfig | None = None,
        kq_conf: QuantizerConfig | None = None,
        rkq_conf: QuantizerConfig | None = None,
        bq_conf: QuantizerConfig | None = None,
        oq_conf: QuantizerConfig | None = None,
        rhq_conf: QuantizerConfig | None = None,
        parallelization_factor=1,
        enable_oq: bool | None = None,
        enable_iq: bool | None = None,
        enable_ebops: bool | None = None,
        beta0: float | None = None,
        **kwargs,
    ):
        cell = QGRUCell(
            units,
            activation=activation,
            recurrent_activation=recurrent_activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint,
            bias_constraint=bias_constraint,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            reset_after=reset_after,
            dtype=kwargs.get('dtype', None),
            trainable=kwargs.get('trainable', True),
            name='qgru_cell',
            seed=seed,
            standalone=False,
            iq_conf=iq_conf,
            paq_conf=paq_conf,
            praq_conf=praq_conf,
            sq_conf=sq_conf,
            kq_conf=kq_conf,
            rkq_conf=rkq_conf,
            bq_conf=bq_conf,
            oq_conf=oq_conf,
            rhq_conf=rhq_conf,
            enable_iq=enable_iq,
            enable_oq=enable_oq,
            enable_ebops=enable_ebops,
            beta0=beta0,
        )
        super(GRU, self).__init__(
            cell,
            return_sequences=return_sequences,
            return_state=return_state,
            go_backwards=go_backwards,
            stateful=stateful,
            unroll=unroll,  # type: ignore
            activity_regularizer=activity_regularizer,
            **kwargs,
        )
        self.input_spec = InputSpec(ndim=3)
        self.use_cudnn = False
        self.parallelization_factor = parallelization_factor
        self._set_unroll()

    def get_config(self):
        base_config = super().get_config()
        conf = {
            'iq_conf': self.cell.iq.config if self.cell.enable_iq else None,
            'paq_conf': self.cell.paq.config,
            'praq_conf': self.cell.praq.config,
            'sq_conf': self.cell.sq.config,
            'kq_conf': self.cell.kq.config,
            'rkq_conf': self.cell.rkq.config,
            'bq_conf': self.cell.bq.config if self.cell.use_bias else None,
            'oq_conf': self.cell.oq.config if self.cell.enable_oq else None,
            'rhq_conf': self.cell.rhq.config,
        }
        # del base_config['cell']
        return {**base_config, **conf}

    def _compute_ebops(self, shape):
        state_shape = (shape[0], self.cell.units)
        cell_shape = (shape[0], shape[2])  # (batch, features)
        return self.cell._compute_ebops(cell_shape, state_shape) * self.parallelization_factor


QLayerBase.register(QGRU)



import keras
import numpy as np
import pytest
from keras import ops
from keras.initializers import RandomNormal, RandomUniform

from hgq.config import QuantizerConfigScope
from hgq.layers import QBatchNormalization
from tests.base import LayerTestBase


class TestBatchNorm(LayerTestBase):
    layer_cls = QBatchNormalization

    @pytest.fixture(params=[2])
    def axis(self, request) -> int:
        return request.param

    @pytest.fixture(params=[True])
    def scale(self, request) -> bool:
        return request.param

    @pytest.fixture(params=[True])
    def center(self, request) -> bool:
        return request.param

    @pytest.fixture
    def layer_kwargs(self, axis, scale, center):
        return {
            'axis': axis,
            'scale': scale,
            'center': center,
            'momentum': 0.99,
            'synchronized': False,
            'moving_mean_initializer': RandomNormal(stddev=1.0),
            'moving_variance_initializer': RandomUniform(minval=0.0, maxval=8.0),
        }

    @pytest.fixture
    def input_shapes(self):
        return (6, 8)

    def test_behavior(self, input_data, layer_kwargs):
        with QuantizerConfigScope(default_q_type='dummy'):
            bn = QBatchNormalization(**{**layer_kwargs, 'momentum': 0.0})  # type: ignore

        input_data = ops.convert_to_tensor(input_data, dtype=bn.dtype)  # type: ignore
        hgq_output = bn(input_data, training=True)
        hgq_output_test = bn(input_data, training=False)
        mean, var = ops.moments(input_data, axes=layer_kwargs['axis'], keepdims=True)  # type: ignore
        ref_output = (input_data - mean) / ops.sqrt(var + bn.epsilon)  # type: ignore

        hgq_output_np: np.ndarray = ops.convert_to_numpy(hgq_output)  # type: ignore
        ref_output_np: np.ndarray = ops.convert_to_numpy(ref_output)  # type: ignore
        hgq_output_test_np: np.ndarray = ops.convert_to_numpy(hgq_output_test)  # type: ignore

        np.allclose(hgq_output_np, ref_output_np, atol=1e-6)
        np.allclose(hgq_output_test_np, ref_output_np)

    def test_da4ml_conversion(self, model: keras.Model, input_data, overflow_mode: str, temp_directory: str):
        super()._test_da4ml_conversion(
            model=model,
            input_data=input_data,
            overflow_mode=overflow_mode,
            temp_directory=temp_directory,
        )



import os
from collections.abc import Sequence
from typing import Any

import keras
import numpy as np
import pytest
from hls4ml.converters import convert_from_keras_model
from keras import ops

from hgq.config import LayerConfigScope, QuantizerConfigScope
from hgq.layers import QLayerBase
from hgq.quantizer.internal import FixedPointQuantizerKBI, FixedPointQuantizerKIF
from hgq.utils import trace_minmax


class CtxGlue:
    def __init__(self, *ctxs):
        self.ctxs = ctxs

    def __enter__(self):
        for ctx in self.ctxs:
            ctx.__enter__()

    def __exit__(self, *args):
        for ctx in self.ctxs:
            ctx.__exit__(*args)


def _assert_equal(a: np.ndarray, b: np.ndarray):
    a, b = np.asanyarray(a).ravel(), np.asanyarray(b).ravel()
    # if keras.backend.backend() != 'torch':
    mismatches = np.where(a != b)[0]
    # else:
    #     # Torch has some extra numerical errors by default...
    #     mismatches = np.where(np.abs(a - b) > 1e-6)[0]
    a_sample = a[mismatches[:5]]
    b_sample = b[mismatches[:5]]
    msg = f"""keras - c synth mismatch. {len(mismatches)} out of {len(a)} samples are different
    Sample:
    {a_sample}
    vs
    {b_sample}
    """
    assert len(mismatches) == 0, msg


class LayerTestBase:
    """Base class for testing HGQ layers"""

    layer_cls: type[QLayerBase] = QLayerBase
    custom_objects = {}
    hls4ml_not_supported = False

    @pytest.fixture(params=[True, False])
    def use_parallel_io(self, request) -> bool:
        return request.param

    @pytest.fixture(params=['kif', 'kbi'])
    def q_type(self, request) -> str:
        return request.param

    @pytest.fixture(params=['WRAP', 'SAT'])
    def overflow_mode(self, request) -> str:
        return request.param

    @pytest.fixture(params=['TRN', 'RND'])
    def round_mode(self, request) -> str:
        return request.param

    @pytest.fixture
    def ctx_scope(self, use_parallel_io: bool, q_type: str, overflow_mode: str, round_mode: str):
        heterogeneous_axis, homogeneous_axis = (None, (0,)) if use_parallel_io else ((), None)
        scope_w = QuantizerConfigScope(
            default_q_type=q_type,
            heterogeneous_axis=None,
            homogeneous_axis=(),
            overflow_mode=overflow_mode,
            round_mode=round_mode,
            br=None,
            ir=None,
            fr=None,
        )
        scope_a = QuantizerConfigScope(
            default_q_type=q_type,
            place='datalane',
            heterogeneous_axis=heterogeneous_axis,
            homogeneous_axis=homogeneous_axis,
            overflow_mode=overflow_mode,
            round_mode=round_mode,
        )
        scope_layer = LayerConfigScope(beta0=0.0, enable_ebops=True)
        return CtxGlue(scope_w, scope_a, scope_layer)

    @pytest.fixture
    def layer_kwargs(self, *args, **kwargs) -> dict[str, Any]:
        """Override this to return kwargs needed for layer construction"""
        return {}

    @pytest.fixture
    def input_shapes(self, *args, **kwargs) -> tuple[tuple[int, ...], ...] | tuple[int, ...]:
        """Override this to return test input shape"""
        raise NotImplementedError

    @pytest.fixture
    def input_data(self, input_shapes, N: int = 5000) -> np.ndarray | tuple[np.ndarray, ...]:
        """Generate test input data"""
        if isinstance(input_shapes[0], int):
            return np.random.randn(N, *input_shapes).astype(np.float32).clip(-5, 5)
        dat = tuple(np.random.randn(N, *shape).astype(np.float32).clip(-5, 5) for shape in input_shapes)  # type: ignore
        if len(dat) == 1:
            return dat[0]
        return dat

    @pytest.fixture
    def layer(self, layer_kwargs, ctx_scope):
        """Create layer instance"""
        with ctx_scope:
            return self.layer_cls(**layer_kwargs)

    @pytest.fixture
    def model(self, layer, input_shapes, use_parallel_io):
        """Create test model with the layer"""
        if isinstance(input_shapes[0], int):
            input_shapes = (input_shapes,)
        inputs = [keras.layers.Input(shape=shape) for shape in input_shapes]
        if len(inputs) == 1:
            inputs = inputs[0]
        outputs = layer(inputs)
        model = keras.Model(inputs, outputs)

        self.perturbe_bw(use_parallel_io, model)
        return model

    def perturbe_bw(self, use_parallel_io, model):
        if use_parallel_io:
            for _layer in model._flatten_layers(False):
                if isinstance(_layer, FixedPointQuantizerKBI):
                    b = np.random.randint(0, 5, _layer._b.shape)
                    i = ops.convert_to_numpy(ops.stop_gradient(_layer.i))
                    b = np.minimum(b, 12 - i)
                    if np.all(b == 0):
                        b.ravel()[0] = 1
                    _layer._b.assign(ops.array(b))
                if isinstance(_layer, FixedPointQuantizerKIF):
                    f = np.random.randint(2, 5, _layer._f.shape)
                    i = ops.convert_to_numpy(ops.stop_gradient(_layer.i))
                    f = np.minimum(f, 12 - i)
                    if np.all(i + f == 0):
                        f.ravel()[0] = 1
                    _layer._f.assign(ops.array(f))
        for _layer in model._flatten_layers(False):
            # Randomize activation values
            if hasattr(_layer, 'bias') and isinstance(_layer.bias, keras.Variable):
                bias = np.random.randn(*_layer.bias.shape)
                _layer.bias.assign(ops.array(bias))

    @pytest.mark.parametrize('format', ['h5', 'keras'])
    def test_serialization(self, model, input_data, format: str, temp_directory: str):
        """Test layer serialization/deserialization"""

        model.compile(optimizer='adam', loss='mean_squared_error')
        original_output = ops.stop_gradient(model(input_data))

        save_path = f'{temp_directory}_model.{format}'
        model.save(save_path)
        loaded_model = keras.models.load_model(save_path, custom_objects=self.custom_objects)
        loaded_output = ops.stop_gradient(loaded_model(input_data))  # type: ignore

        original_output = ops.convert_to_numpy(original_output)
        loaded_output = ops.convert_to_numpy(loaded_output)

        np.testing.assert_array_equal(original_output, loaded_output)  # type: ignore

        os.system(f"rm '{save_path}'")

    def test_hls4ml_conversion(self, model: keras.Model, input_data, temp_directory: str, use_parallel_io: bool, q_type: str):
        if self.hls4ml_not_supported:
            pytest.skip('No synth test')
        """Test hls4ml conversion and bit-exactness"""

        trace_keras_output_0 = trace_minmax(model, input_data, return_results=True, verbose=2)

        hls_model = convert_from_keras_model(
            model,
            output_dir=f'{temp_directory}/hls4ml_prj',
            backend='Vitis',
            io_type='io_parallel' if use_parallel_io else 'io_stream',
            hls_config={'Model': {'Precision': 'ap_fixed<1,0>', 'ReuseFactor': 1, 'Strategy': 'distributed_arithmetic'}},
        )

        hls_model.compile()
        if q_type == 'kif':
            keras_output_0 = model.predict(input_data, batch_size=5000)
            self.assert_equal(keras_output_0, trace_keras_output_0)

        if isinstance(input_data, Sequence):
            input_data = tuple(_input_data * 2 for _input_data in input_data)  # type: ignore
        else:
            input_data = input_data * 2  # Test for overflow in the same time

        keras_output = model.predict(input_data, batch_size=5000)
        hls_output: np.ndarray = hls_model.predict(input_data).reshape(keras_output.shape)  # type: ignore
        self.assert_equal(keras_output, hls_output)

        os.system(f"rm -rf '{temp_directory}/hls4ml_prj'")

    def _test_da4ml_conversion(self, model: keras.Model, input_data, overflow_mode: str, temp_directory: str):
        if overflow_mode == 'SAT':
            pytest.skip()

        from da4ml.codegen import HLSModel, VerilogModel
        from da4ml.converter import trace_model
        from da4ml.trace import comb_trace

        inp, out = trace_model(model)
        comb = comb_trace(inp, out)
        verilog_model = VerilogModel(comb, 'test', f'{temp_directory}/verilog_proj', 5)
        hls_model = HLSModel(comb, 'test', f'{temp_directory}/hls_proj')
        if all(qint.min == qint.max for qint in verilog_model._pipe.out_qint):  # type: ignore
            return  # No need to synthesize as model is trivial (all outputs are constant)
        verilog_model.compile(nproc=1, openmp=False)
        hls_model.compile(openmp=False)
        keras_output = model.predict(input_data, batch_size=5000)
        if isinstance(keras_output, Sequence):
            keras_output = tuple(_out.reshape(np.shape(_out)[0], -1) for _out in keras_output)
        else:
            keras_output = keras_output.reshape(np.shape(keras_output)[0], -1)
        if isinstance(input_data, np.ndarray):
            v_input_data = input_data.reshape(np.shape(input_data)[0], -1)
        else:
            v_input_data = np.concatenate([_inp.reshape(np.shape(_inp)[0], -1) for _inp in input_data], axis=0)
        v_output = verilog_model.predict(v_input_data)
        h_output = hls_model.predict(v_input_data)

        np.testing.assert_array_equal(v_output, h_output)

        self.assert_equal(keras_output, v_output)
        os.system(f"rm -rf '{temp_directory}/verilog_proj'")
        os.system(f"rm -rf '{temp_directory}/hls_proj'")

    def assert_equal(self, keras_output, hls_output):
        _assert_equal(keras_output, hls_output)

    def test_training(self, model: keras.Model, input_data, overflow_mode: str, *args, **kwargs):
        """Test basic training step"""

        model_wrap = keras.Sequential([model, keras.layers.Flatten(), keras.layers.Lambda(lambda x: ops.sum(x))])
        if isinstance(input_data, Sequence):
            input_data = tuple(_input_data for _input_data in input_data)  # type: ignore
        else:
            input_data = ops.convert_to_tensor(input_data, dtype=model.dtype)  # type: ignore

        initial_weights_np = [w.numpy() for w in model.trainable_variables]

        opt = keras.optimizers.SGD()
        loss = keras.losses.MeanAbsoluteError()
        model(input_data, training=True)  # Adapt init bitwidth

        data_len = len(input_data[0]) if isinstance(input_data, Sequence) else len(input_data)
        shape = (data_len, *model.output.shape[1:])  # type: ignore
        labels = ops.array(np.random.rand(*shape), dtype='float32')
        model_wrap.compile(optimizer=opt, loss=loss)  # type: ignore
        model_wrap.train_on_batch(input_data, labels)

        trained_weights = [w for w in model.trainable_variables]

        boom = []
        for w0, w1 in zip(initial_weights_np, trained_weights):
            if w1.name in 'bif':
                continue
            if np.prod(w1.shape) < 10 and 'SAT' in overflow_mode:
                # Overflowing weight doesn't receive grad in SAT mode
                # Chance of all overflow is high for small-sized weights, skip them
                continue
            if np.array_equal(w0, w1.numpy()):
                boom.append(f'{w1.path}')
        assert not boom, f'Weight {" AND ".join(boom)} did not change'
        assert any(np.any(w0 != w1.numpy()) for w0, w1 in zip(initial_weights_np, trained_weights) if w1.name in 'bif')



import numpy as np
from keras import ops
from keras.src import constraints, initializers, regularizers
from keras.src.backend.config import epsilon
from keras.src.layers.core.einsum_dense import _analyze_einsum_string

from ..quantizer.config import QuantizerConfig
from .core.einsum_dense import QEinsumDense


class QEinsumDenseBatchnorm(QEinsumDense):  # type: ignore
    def __init__(
        self,
        equation,
        output_shape,
        activation=None,
        bias_axes=None,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        kq_conf: None | QuantizerConfig = None,
        iq_conf: None | QuantizerConfig = None,
        bq_conf: None | QuantizerConfig = None,
        normalize_axes: str | None = None,
        momentum=0.99,
        epsilon=1e-3,
        center=True,
        scale=True,
        beta_initializer='zeros',
        gamma_initializer='ones',
        moving_mean_initializer='zeros',
        moving_variance_initializer='ones',
        beta_regularizer=None,
        gamma_regularizer=None,
        beta_constraint=None,
        gamma_constraint=None,
        synchronized=False,
        **kwargs,
    ):
        super().__init__(
            equation=equation,
            output_shape=output_shape,
            activation=activation,
            bias_axes=bias_axes,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            iq_conf=iq_conf,
            kq_conf=kq_conf,
            bq_conf=bq_conf,
            **kwargs,
        )
        self.synchronized = synchronized
        self.momentum = float(momentum)
        self.epsilon = float(epsilon)
        self.center = center
        self.scale = scale
        self.beta_initializer = initializers.get(beta_initializer)
        self.gamma_initializer = initializers.get(gamma_initializer)
        self.moving_mean_initializer = initializers.get(moving_mean_initializer)
        self.moving_variance_initializer = initializers.get(
            moving_variance_initializer,
        )
        self.beta_regularizer = regularizers.get(beta_regularizer)
        self.gamma_regularizer = regularizers.get(gamma_regularizer)
        self.beta_constraint = constraints.get(beta_constraint)
        self.gamma_constraint = constraints.get(gamma_constraint)
        self.supports_masking = False

        normalize_axes = normalize_axes or bias_axes
        assert normalize_axes is not None, 'Either normalize_axes or bias_axes must be provided.'
        self.normalize_axes = normalize_axes
        self._check_normalize_axes()

    def _check_normalize_axes(self):
        kernel_axes, output_axes = self.equation.split(',')[1].split('->')
        for c in self.normalize_axes:
            if c not in kernel_axes:
                raise ValueError(f'Axis {c} not found in kernel axes {kernel_axes}')
            if c not in output_axes:
                raise ValueError(f'Axis {c} not found in output axes {output_axes}')
            if self.bias_axes and c not in self.bias_axes:
                raise ValueError(f'Axis {c} not found in bias axes {self.bias_axes}')

    @property
    def kq(self):
        return self._kq

    @property
    def bq(self):
        return self._bq

    def _compute_fused_einsum_specs(self, input_shape):
        # Replace '...' with '0' to match the behavior of keras
        equation = self.equation.replace('...', '0')
        kernel_axes, output_axes = equation.split(',')[1].split('->')
        shape_data = _analyze_einsum_string(equation, self.normalize_axes, input_shape, self.partial_output_shape)
        kernel_shape, _norm_bias_cast_shape, output_shape = shape_data
        assert _norm_bias_cast_shape is not None
        if '0' in output_axes:
            _pad_len = len(output_shape) - len(output_axes) + 1
            output_axes = output_axes.replace('0', '.' * _pad_len)

        assert len(output_axes) == len(output_shape), f'{output_axes} != {output_shape}'
        # Order normalize_axes to match kernel_axes
        normalize_axes = ''.join(c for c in kernel_axes if c in self.normalize_axes)

        _reduction_axis = tuple(i for i, c in enumerate(output_axes) if c not in normalize_axes)
        norm_shape = tuple(output_shape[i] for i, c in enumerate(output_axes) if c in normalize_axes)

        assert np.prod(norm_shape) == np.prod(_norm_bias_cast_shape), f'{norm_shape} != {_norm_bias_cast_shape}'

        fused_kernel_equation = f'{kernel_axes},{normalize_axes}->{kernel_axes}'

        self._reduction_axes = _reduction_axis
        self._norm_offset_cast_shape = _norm_bias_cast_shape
        self.fused_kernel_equation = fused_kernel_equation
        return norm_shape

    def build(self, input_shape):
        super().build(input_shape)
        norm_shape = self._compute_fused_einsum_specs(input_shape)

        if self.scale:
            self.bn_gamma = self.add_weight(
                shape=norm_shape,
                name='gamma',
                initializer=self.gamma_initializer,
                regularizer=self.gamma_regularizer,
                constraint=self.gamma_constraint,
                trainable=True,
                autocast=False,
            )
        assert self.bias is not None
        self.bias.trainable = self.center

        self.moving_mean = self.add_weight(
            shape=norm_shape,
            name='moving_mean',
            initializer=self.moving_mean_initializer,
            trainable=False,
            autocast=False,
        )
        self.moving_variance = self.add_weight(
            shape=norm_shape,
            name='moving_variance',
            initializer=self.moving_variance_initializer,
            trainable=False,
            autocast=False,
        )

    def get_fused_qkernel_and_qbias(self, training, mean, var):
        assert self.bias is not None
        assert self.bq is not None
        if self.scale:
            bn_gamma = self.bn_gamma
        else:
            bn_gamma = 1
        scaler = bn_gamma / ops.sqrt(var + epsilon())  # type: ignore
        kernel = self.kernel
        fused_kernel = ops.einsum(self.fused_kernel_equation, kernel, scaler)
        fused_qkernel = self.kq(fused_kernel, training=training)

        offset = -mean * scaler
        offset = ops.reshape(offset, self._norm_offset_cast_shape)
        fused_qbias = self.bq(self.bias + offset, training=training)
        return fused_qkernel, fused_qbias

    @property
    def qkernel(self):
        mean, var = self.moving_mean, self.moving_variance
        return self.get_fused_qkernel_and_qbias(training=False, mean=mean, var=var)[0]

    @property
    def qbias(self):
        mean, var = self.moving_mean, self.moving_variance
        return self.get_fused_qkernel_and_qbias(training=False, mean=mean, var=var)[1]

    def call(self, inputs, training=None):  # type: ignore
        if self.enable_iq:
            inputs = self.iq(inputs, training=training)

        if training and self.trainable:
            x = ops.einsum(self.equation, inputs, self.kernel)
            mean, var = ops.moments(
                x,
                self._reduction_axes,
                keepdims=False,
                synchronized=self.synchronized,
            )  # type: ignore
            self.moving_mean.assign(
                self.moving_mean * self.momentum + mean * (1.0 - self.momentum),
            )
            self.moving_variance.assign(
                self.moving_variance * self.momentum + var * (1.0 - self.momentum),
            )
        else:
            var, mean = self.moving_variance, self.moving_mean

        qkernel, qbias = self.get_fused_qkernel_and_qbias(training, mean, var)

        x = ops.einsum(self.equation, inputs, qkernel) + qbias
        if self.activation is not None:
            x = self.activation(x)
        return x



import keras
import numpy as np
import pytest
from keras import layers, ops

from hgq.config import QuantizerConfig, QuantizerConfigScope
from hgq.layers.rnn import QGRU, QSimpleRNN
from hgq.layers.rnn.simple_rnn import QRNN
from tests.base import LayerTestBase


class RNNTestBase(LayerTestBase):
    layer_cls = QRNN
    keras_layer_cls = layers.RNN

    @pytest.fixture(params=((5, 8), (31, 7)))
    def input_shapes(self, request):
        return request.param

    @pytest.fixture
    def layer_kwargs(self, pool_size, strides, padding):
        return {
            'pool_size': pool_size,
            'strides': strides,
            'padding': padding,
        }

    def test_hls4ml_conversion(
        self, model: keras.Model, input_data: np.ndarray, temp_directory: str, use_parallel_io: bool, q_type: str
    ):
        pytest.skip('hls4ml support not yet ready')

    def test_behavior(self, input_data, layer_kwargs):
        raise NotImplementedError()


class TestQSimpleRNN(RNNTestBase):
    layer_cls = QSimpleRNN
    keras_layer_cls = layers.SimpleRNN

    @pytest.fixture(params=[9])
    def units(self, request):
        return request.param

    @pytest.fixture(params=['linear'])
    def activation(self, request):
        return request.param

    @pytest.fixture(params=[True, False])
    def layer_kwargs(self, units, activation, request):
        return {
            'units': units,
            'activation': activation,
            'return_sequences': request.param,
            'use_bias': not request.param,
            'go_backwards': request.param,
        }

    def test_behavior(self, input_data, layer_kwargs):
        layer_kwargs = layer_kwargs.copy()
        layer_kwargs['activation'] = 'tanh'

        keras_layer = self.keras_layer_cls(**layer_kwargs)
        with QuantizerConfigScope(default_q_type='dummy'):
            q_layer = self.layer_cls(**layer_kwargs, enable_ebops=False, paq_conf=QuantizerConfig('dummy'))

        keras_layer.build(input_data.shape)
        q_layer.build(input_data.shape)

        for w0, w1 in zip(keras_layer.weights, q_layer.weights):
            w1.assign(w0)
            assert w0.name == w1.name
        assert len(keras_layer.weights) == len(q_layer.weights)

        keras_output = keras_layer(input_data)
        q_output = q_layer(input_data)

        assert ops.all(keras_output == q_output)


class TestQGRU(RNNTestBase):
    layer_cls = QGRU
    keras_layer_cls = layers.GRU

    @pytest.fixture(params=[9])
    def units(self, request):
        return request.param

    @pytest.fixture(params=['linear'])
    def activation(self, request):
        return request.param

    @pytest.fixture(params=['linear'])
    def recurrent_activation(self, request):
        return request.param

    @pytest.fixture(params=[True, False])
    def layer_kwargs(self, units, activation, recurrent_activation, request):
        return {
            'units': units,
            'activation': activation,
            'recurrent_activation': recurrent_activation,
            'return_sequences': request.param,
            'use_bias': not request.param,
            'go_backwards': request.param,
        }

    def test_behavior(self, input_data, layer_kwargs):
        layer_kwargs['activation'] = 'tanh'
        layer_kwargs['recurrent_activation'] = 'sigmoid'

        keras_layer = self.keras_layer_cls(**layer_kwargs)
        with QuantizerConfigScope(default_q_type='dummy'):
            q_layer = self.layer_cls(
                **layer_kwargs, enable_ebops=False, paq_conf=QuantizerConfig('dummy'), praq_conf=QuantizerConfig('dummy')
            )

        keras_layer.build(input_data.shape)
        q_layer.build(input_data.shape)

        for w0, w1 in zip(keras_layer.weights, q_layer.weights):
            w1.assign(w0)
            assert w0.name == w1.name
        assert len(keras_layer.weights) == len(q_layer.weights)

        keras_output = keras_layer(input_data)
        q_output = q_layer(input_data)

        assert ops.all(keras_output == q_output), f'{keras_output} != {q_output}'



from .base import BitwidthMapperBase, DefaultBitwidthMapper, DummyQuantizer, TrainableQuantizerBase, numbers
from .fixed_point_quantizer import FixedPointQuantizerBase, FixedPointQuantizerKBI, FixedPointQuantizerKIF
from .float_point_quantizer import FloatPointQuantizer

__all__ = [
    'BitwidthMapperBase',
    'DefaultBitwidthMapper',
    'DummyQuantizer',
    'TrainableQuantizerBase',
    'FixedPointQuantizerBase',
    'FixedPointQuantizerKBI',
    'FixedPointQuantizerKIF',
    'FloatPointQuantizer',
    'numbers',
]



import numpy as np
import pytest
from keras import ops

from hgq.config import QuantizerConfigScope
from hgq.layers import QSoftmax, QUnaryFunctionLUT
from tests.base import LayerTestBase


def custom_fn(x):
    return ops.tanh(x) - ops.sin(x) + ops.log(ops.abs(x) + 1)  # type: ignore


class TestQUnaryFunctionLUT(LayerTestBase):
    layer_cls = QUnaryFunctionLUT
    custom_objects = {'custom_fn': custom_fn}

    @pytest.fixture(params=['sigmoid', 'tanh', custom_fn])
    def layer_kwargs(self, request):
        return {'activation': request.param}

    @pytest.fixture
    def input_shapes(self):
        return (8,)


class TestSoftmax(LayerTestBase):
    layer_cls = QSoftmax

    @pytest.fixture(params=[-1, (-2, -1)])
    def axis(self, request):  # return request.param
        return request.param

    @pytest.fixture(params=[(8,), (4, 4)])
    def input_shapes(self, request, axis):
        if axis == (-2, -1) and request.param != (4, 4):
            pytest.skip('Only support 2D input with axis=(-2, -1)')
        return request.param

    @pytest.fixture(params=[True, False])
    def stable(self, request):
        return request.param

    @pytest.fixture
    def layer_kwargs(self, axis, stable, use_parallel_io):
        if not use_parallel_io and axis != -1:
            pytest.skip('hls4ml only support axis=-1 with io_stream')

        return {
            'axis': axis,
            'stable': stable,
        }

    def test_behavior(self, input_data, layer_kwargs):
        with QuantizerConfigScope(default_q_type='dummy'):
            softmax = QSoftmax(**layer_kwargs)

        axis = layer_kwargs['axis']
        hgq_output = softmax(input_data)
        if axis == -1:
            ref_output = ops.nn.softmax(input_data, axis=-1)
        else:
            shape = input_data.shape
            input_data = ops.reshape(input_data, shape[:-2] + (-1,))
            ref_output = ops.reshape(ops.nn.softmax(input_data, axis=-1), shape)

        hgq_output_np: np.ndarray = ops.convert_to_numpy(hgq_output)  # type: ignore
        ref_output_np: np.ndarray = ops.convert_to_numpy(ref_output)  # type: ignore

        np.testing.assert_allclose(hgq_output_np, ref_output_np, atol=1e-6)



from .base import QLayerBase, QLayerBaseMultiInputs, QLayerBaseSingleInput
from .dense import QBatchNormDense, QDense
from .einsum_dense import QEinsumDense

__all__ = [
    'QLayerBaseSingleInput',
    'QLayerBase',
    'QLayerBaseMultiInputs',
    'QEinsumDense',
    'QDense',
    'QBatchNormDense',
]



"""HGQ2/QKeras
QKeras compatibility layer for HGQ2. This API is for compatibility purpose only, and only exposes a minimal subset of the HGQ2's functionality. It is not recommended for new projects.

If you are starting a new project, please use the HGQ2 API directly for more flexibility and control.
"""

from .quantizers import *



from warnings import warn

import keras
from keras import ops
from keras.initializers import Constant
from keras.saving import deserialize_keras_object, register_keras_serializable
from keras.src.layers.input_spec import InputSpec
from keras.src.layers.rnn.simple_rnn import RNN, SimpleRNN, SimpleRNNCell

from ...config import HardTanhConfig, QuantizerConfig
from ...layers.core.base import QLayerBase, QLayerMeta
from ...quantizer import Quantizer
from ...quantizer.internal import FixedPointQuantizerBase
from ..core.base import QLayerBaseSingleInput


class QSimpleRNNCell(QLayerBaseSingleInput, SimpleRNNCell):
    """Cell class for the QSimpleRNN layer.

    This class processes one step within the whole time sequence input, whereas
    `QSimpleRNN` processes the whole sequence.

    Parameters
    ----------
    units : int
        Positive integer, dimensionality of the output space.
    activation : str, optional
        Activation function to use. Default: "linear".
    use_bias : bool, optional
        Whether the layer should use a bias vector. Default: True.
    kernel_initializer : str, optional
        Initializer for the `kernel` weights matrix,
        used for the linear transformation of the inputs. Default: "glorot_uniform".
    recurrent_initializer : str, optional
        Initializer for the `recurrent_kernel` weights matrix, used for the linear transformation
        of the recurrent state. Default: "orthogonal".
    bias_initializer : str, optional
        Initializer for the bias vector. Default: "zeros".
    kernel_regularizer : optional
        Regularizer function applied to the `kernel` weights matrix. Default: None.
    recurrent_regularizer : optional
        Regularizer function applied to the `recurrent_kernel` weights matrix. Default: None.
    bias_regularizer : optional
        Regularizer function applied to the bias vector. Default: None.
    kernel_constraint : optional
        Constraint function applied to the `kernel` weights matrix. Default: None.
    recurrent_constraint : optional
        Constraint function applied to the `recurrent_kernel` weights matrix. Default: None.
    bias_constraint : optional
        Constraint function applied to the bias vector. Default: None.
    dropout : float, optional
        Float between 0 and 1. Fraction of the units to drop for the
        linear transformation of the inputs. Default: 0.
    recurrent_dropout : float, optional
        Float between 0 and 1. Fraction of the units to drop for the
        linear transformation of the recurrent state. Default: 0.
    seed : int, optional
        Random seed for dropout.
    enable_sq : bool or None, optional
        Whether to enable state quantizer.
        When the output is already quantized and the state is 0-inited,
        the state quantizer should be disabled to avoid double quantization.
        Default: None, which means it will be set to the value of `standalone`.
    iq_conf : QuantizerConfig or None, optional
        Quantizer configuration for input quantizer. Default: None.
    sq_conf : QuantizerConfig or None, optional
        Quantizer configuration for state quantizer. Default: None.
    kq_conf : QuantizerConfig or None, optional
        Quantizer configuration for kernel quantizer. Default: None.
    rkq_conf : QuantizerConfig or None, optional
        Quantizer configuration for recurrent kernel quantizer. Default: None.
    bq_conf : QuantizerConfig or None, optional
        Quantizer configuration for bias quantizer. Default: None.
    paq_conf : QuantizerConfig or None, optional
        Quantizer configuration for pre-activation quantizer. Default: None.
    standalone : bool, optional
        Whether this cell is used standalone or as part of a larger RNN layer. Default: True.
    """

    @property
    def kq(self):
        "Kernel Quantizer"
        return self._kq

    @property
    def rkq(self):
        "Recurrent Kernel Quantizer"
        return self._rkq

    @property
    def iq(self):
        "Input Quantizer"
        return self._iq

    @property
    def sq(self):
        "State Quantizer"
        if not self.enable_sq:
            raise ValueError('State Quantizer is not enabled.')
        return self._sq

    @property
    def bq(self):
        "Bias Quantizer"
        if not self.use_bias:
            return None
        return self._bq

    @property
    def paq(self):
        "Pre-Activation Quantizer"
        return self._paq

    @property
    def qkernel(self):
        return self.kq(self.kernel)

    @property
    def qrecurrent_kernel(self):
        return self.rkq(self.recurrent_kernel)

    @property
    def qbias(self):
        if not self.use_bias:
            return None
        return self.bq(self.bias)  # type: ignore

    @property
    def enable_sq(self):
        return self._enable_sq

    def __init__(
        self,
        units,
        activation='linear',
        use_bias=True,
        kernel_initializer='glorot_uniform',
        recurrent_initializer='orthogonal',
        bias_initializer='zeros',
        kernel_regularizer=None,
        recurrent_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        recurrent_constraint=None,
        bias_constraint=None,
        dropout=0.0,
        recurrent_dropout=0.0,
        seed=None,
        enable_sq: bool | None = None,
        iq_conf: QuantizerConfig | None = None,
        sq_conf: QuantizerConfig | None = None,
        kq_conf: QuantizerConfig | None = None,
        rkq_conf: QuantizerConfig | None = None,
        bq_conf: QuantizerConfig | None = None,
        paq_conf: QuantizerConfig | None = None,
        standalone=True,
        **kwargs,
    ):
        super().__init__(
            units=units,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint,
            bias_constraint=bias_constraint,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            iq_conf=iq_conf,
            seed=seed,
            **kwargs,
        )

        if enable_sq is None:
            enable_sq = standalone
        self._enable_sq = enable_sq

        iq_conf = iq_conf or QuantizerConfig(place='datalane')
        sq_conf = sq_conf or QuantizerConfig(place='datalane')
        kq_conf = kq_conf or QuantizerConfig(place='weight')
        rkq_conf = rkq_conf or QuantizerConfig(place='weight')
        bq_conf = bq_conf or QuantizerConfig(place='bias')
        paq_conf = paq_conf or HardTanhConfig(place='datalane')

        if self.enable_sq:
            self._sq = Quantizer(sq_conf, name='sq')
        self._kq = Quantizer(kq_conf, name='kq')
        self._rkq = Quantizer(rkq_conf, name='rkq')
        if use_bias:
            self._bq = Quantizer(bq_conf, name='bq')
        self._paq = Quantizer(paq_conf, name='paq')

        self.standalone = standalone

    def build(self, input_shape):
        if self.enable_sq:
            self._sq.build((None, self.units))
        self._kq.build((input_shape[-1], self.units))
        self._rkq.build((self.units, self.units))
        if self.use_bias:
            self._bq.build((self.units,))
        self._paq.build((None, self.units))
        super().build(input_shape)

    def call(self, sequence, states, training=False):
        prev_output = states[0] if isinstance(states, (list, tuple)) else states
        dp_mask = self.get_dropout_mask(sequence)
        rec_dp_mask = self.get_recurrent_dropout_mask(prev_output)

        if training and dp_mask is not None:
            sequence = sequence * dp_mask
        if training and rec_dp_mask is not None:
            prev_output = prev_output * rec_dp_mask

        qkernel = self.kq(self.kernel, training=training)
        qrecurrent_kernel = self.rkq(self.recurrent_kernel, training=training)
        if self.enable_sq:
            qstate = self.sq(prev_output, training=training)
        else:
            # If the output of the activation is already quantized, sq shall be disabled.
            qstate = prev_output
        qsequence = self.iq(sequence, training=training)

        h = ops.matmul(qsequence, qkernel)
        if self.bias is not None:
            h += self.bq(self.bias, training=training)  # type: ignore

        output = h + ops.matmul(qstate, qrecurrent_kernel)  # type: ignore

        output = self.paq(output)
        output = self.activation(output)

        new_state = [output] if isinstance(states, (list, tuple)) else output
        return output, new_state

    def get_config(self):
        return {
            'iq_conf': self._iq.config,
            'sq_conf': self._sq.config if self.enable_sq else None,
            'kq_conf': self._kq.config,
            'rkq_conf': self._rkq.config,
            'bq_conf': self._bq.config if self.use_bias else None,
            'paq_conf': self._paq.config,
            'enable_sq': self._enable_sq,
            'standalone': self.standalone,
            **super().get_config(),
        }

    def _compute_ebops(self, shape):
        bw_inp = self.iq.bits_(shape)
        bw_ker = self.kq.bits_(ops.shape(self.kernel))
        ebops1 = ops.sum(ops.matmul(bw_inp, bw_ker))
        if self.enable_sq:
            bw_state = self.sq.bits_((1, self.units))
        else:
            bw_state = self.paq.bits_((1, self.units))
        bw_rker = self.rkq.bits_(ops.shape(self.recurrent_kernel))
        ebops2 = ops.sum(ops.matmul(bw_state, bw_rker))
        ebops = ebops1 + ebops2  # type: ignore
        if self.bq is not None:
            bw_bias = self.bq.bits_(ops.shape(self.bias))
            size = ops.cast(ops.prod(shape), self.dtype)
            ebops = ebops + ops.mean(bw_bias) * size  # type: ignore
        return ebops

    @property
    def enable_ebops(self):
        # When used as a sublayer in the RNN layer, standalone is set to False
        # EBOPs computation handled on the higher level RNN layer
        return self._enable_ebops and self.standalone


def _is_wrap_quantizer(q: Quantizer) -> bool:
    if not isinstance(q, Quantizer):
        return False
    if not isinstance(q.quantizer, FixedPointQuantizerBase):
        return False
    return q.quantizer.overflow_mode == 'WRAP'


@register_keras_serializable(package='hgq')
class QRNN(RNN, metaclass=QLayerMeta):
    __output_quantizer_handled__ = True

    def _set_unroll(self):
        backend = keras.backend.backend()
        has_wrap_quantizers = any(_is_wrap_quantizer(layer) for layer in self._flatten_layers())
        if backend == 'jax' and has_wrap_quantizers:
            # JAX tracer issues when using jax scan with WRAP quantizers (range update is side effect)
            # Force unrolling in this case
            if self.unroll is False:
                warn('JAX backend does not support WRAP quantizers with rolled RNNs. Forcing unrolling.')
            self.unroll = True
        if self.unroll is None:
            self.unroll = False  # Keras default

    def build(self, sequences_shape, initial_state_shape=None):
        seq_len = sequences_shape[1]
        if self.parallelization_factor == -1:
            self.parallelization_factor = seq_len

        if self.enable_ebops:
            self._beta = self.add_weight(
                name='beta',
                shape=(),
                initializer=self.cell._beta0,
                trainable=False,
            )
            self._ebops = self.add_weight(
                name='ebops',
                shape=(),
                initializer=Constant(0.0),
                trainable=False,
                dtype='uint32',
            )
        else:
            self._beta = None
            self._ebops = None

        super().build(sequences_shape, initial_state_shape)

    def get_config(self):
        return {
            'parallelization_factor': self.parallelization_factor,
            'enable_ebops': self.enable_ebops,
            'enable_iq': self.enable_iq,
            'enable_oq': self.enable_oq,
            'beta0': self.cell._beta0,
            **super().get_config(),
        }

    @property
    def enable_ebops(self):
        return self.cell._enable_ebops

    @property
    def enable_iq(self):
        return self.cell._enable_iq

    @property
    def enable_oq(self):
        return self.cell._enable_oq

    @property
    def beta(self):
        if self._beta is None:
            return ops.cast(0, 'float32')
        return ops.cast(self._beta, ops.dtype(self._beta))

    @property
    def ebops(self):
        if self._ebops is None:
            return ops.cast(0, 'uint32')
        return ops.cast(self._ebops, ops.dtype(self._ebops))

    @classmethod
    def from_config(cls, config, custom_objects=None):
        config = deserialize_keras_object(config, custom_objects=custom_objects)
        return super().from_config(config)


@register_keras_serializable(package='hgq')
class QSimpleRNN(QRNN, SimpleRNN):
    """Quantized Fully-connected RNN where the output is to be fed back as the new input.

    When the jax backend is used, if any `WRAP` quantizers are used, unroll will
    be set to `True` to avoid the side effect issue in the `jax.lax.scan` loop.

    Parameters
    ----------
    units : int
        Positive integer, dimensionality of the output space.
    activation : str, optional
        Activation function to use.
        Default: linear, effectively hard_tanh by the pre-activation quantizer.
    use_bias : bool, optional
        Whether the layer uses a bias vector. Default: True.
    kernel_initializer : str, optional
        Initializer for the `kernel` weights matrix,
        used for the linear transformation of the inputs. Default: "glorot_uniform".
    recurrent_initializer : str, optional
        Initializer for the `recurrent_kernel` weights matrix, used for the linear transformation
        of the recurrent state. Default: "orthogonal".
    bias_initializer : str, optional
        Initializer for the bias vector. Default: "zeros".
    kernel_regularizer : optional
        Regularizer function applied to the `kernel` weights matrix. Default: None.
    recurrent_regularizer : optional
        Regularizer function applied to the `recurrent_kernel` weights matrix. Default: None.
    bias_regularizer : optional
        Regularizer function applied to the bias vector. Default: None.
    activity_regularizer : optional
        Regularizer function applied to the output of the layer (its "activation"). Default: None.
    kernel_constraint : optional
        Constraint function applied to the `kernel` weights matrix. Default: None.
    recurrent_constraint : optional
        Constraint function applied to the `recurrent_kernel` weights matrix. Default: None.
    bias_constraint : optional
        Constraint function applied to the bias vector. Default: None.
    dropout : float, optional
        Float between 0 and 1. Fraction of the units to drop for the linear transformation
        of the inputs. Default: 0.
    recurrent_dropout : float, optional
        Float between 0 and 1. Fraction of the units to drop for the linear transformation of the
        recurrent state. Default: 0.
    return_sequences : bool, optional
        Whether to return the last output in the output sequence, or the full sequence. Default: False.
    return_state : bool, optional
        Whether to return the last state in addition to the output. Default: False.
    go_backwards : bool, optional
        If True, process the input sequence backwards and return the reversed sequence. Default: False.
    stateful : bool, optional
        If True, the last state for each sample at index i in a batch will be used as initial
        state for the sample of index i in the following batch. Default: False.
    unroll : bool or None, optional
        None is equivalent to False. However, for the JAX backend, if
        any `WRAP` quantizers are used, unroll will be set to True
        to avoid the side effect issue in the `jax.lax.scan` loop.
        If True, the network will be unrolled,
        else a symbolic loop will be used.
        Unrolling can speed-up a RNN,
        although it tends to be more memory-intensive.
        Unrolling is only suitable for short sequences. Default: None.
    seed : int, optional
        Random seed for dropout.
    iq_conf : QuantizerConfig or None, optional
        Quantizer configuration for input quantizer. Default: None.
    sq_conf : QuantizerConfig or None, optional
        Quantizer configuration for state quantizer. Default: None.
    kq_conf : QuantizerConfig or None, optional
        Quantizer configuration for kernel quantizer. Default: None.
    rkq_conf : QuantizerConfig or None, optional
        Quantizer configuration for recurrent kernel quantizer. Default: None.
    bq_conf : QuantizerConfig or None, optional
        Quantizer configuration for bias quantizer. Default: None.
    oq_conf : QuantizerConfig or None, optional
        Quantizer configuration for output quantizer. Default: None.
    paq_conf : QuantizerConfig or None, optional
        Quantizer configuration for pre-activation quantizer. Default: None.
    parallelization_factor : int, optional
        Factor for parallelization. Default: -1 (automatic).
    enable_sq : bool or None, optional
        Whether to enable state quantizer.
        When the output is already quantized and the state is 0-inited,
        the state quantizer should be disabled to avoid double quantization.
        Default: False.
    enable_oq : bool or None, optional
        Whether to enable output quantizer. Default: None.
    enable_iq : bool or None, optional
        Whether to enable input quantizer. Default: None.
    enable_ebops : bool or None, optional
        Whether to enable energy-efficient bit operations. Default: None.
    beta0 : float or None, optional
        Beta0 parameter for quantizer. Default: None.

    Notes
    -----
    sequence : array_like
        A 3D tensor, with shape `[batch, timesteps, feature]`.
    mask : array_like, optional
        Binary tensor of shape `[batch, timesteps]` indicating whether
        a given timestep should be masked. An individual `True` entry
        indicates that the corresponding timestep should be utilized,
        while a `False` entry indicates that the corresponding timestep
        should be ignored.
    training : bool, optional
        Python boolean indicating whether the layer should behave in
        training mode or in inference mode.
        This argument is passed to the cell when calling it.
        This is only relevant if `dropout` or `recurrent_dropout` is used.
    initial_state : list, optional
        List of initial state tensors to be passed to the first
        call of the cell.
    """

    def __init__(
        self,
        units,
        activation='linear',
        use_bias=True,
        kernel_initializer='glorot_uniform',
        recurrent_initializer='orthogonal',
        bias_initializer='zeros',
        kernel_regularizer=None,
        recurrent_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        recurrent_constraint=None,
        bias_constraint=None,
        dropout=0.0,
        recurrent_dropout=0.0,
        return_sequences=False,
        return_state=False,
        go_backwards=False,
        stateful=False,
        unroll=None,
        seed=None,
        iq_conf: QuantizerConfig | None = None,
        sq_conf: QuantizerConfig | None = None,
        kq_conf: QuantizerConfig | None = None,
        rkq_conf: QuantizerConfig | None = None,
        bq_conf: QuantizerConfig | None = None,
        oq_conf: QuantizerConfig | None = None,
        paq_conf: QuantizerConfig | None = None,
        parallelization_factor=-1,
        enable_oq: bool | None = None,
        enable_iq: bool | None = None,
        enable_sq: bool | None = False,
        enable_ebops: bool | None = None,
        beta0: float | None = None,
        **kwargs,
    ):
        cell = QSimpleRNNCell(
            units,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint,
            bias_constraint=bias_constraint,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            seed=seed,
            dtype=kwargs.get('dtype', None),
            trainable=kwargs.get('trainable', True),
            name='simple_rnn_cell',
            iq_conf=iq_conf,
            sq_conf=sq_conf,
            kq_conf=kq_conf,
            rkq_conf=rkq_conf,
            bq_conf=bq_conf,
            oq_conf=oq_conf,
            enable_sq=enable_sq,
            paq_conf=paq_conf,
            standalone=False,
            enable_oq=enable_oq,
            enable_iq=enable_iq,
            enable_ebops=enable_ebops,
            beta0=beta0,
        )
        super(SimpleRNN, self).__init__(
            cell,
            return_sequences=return_sequences,
            return_state=return_state,
            go_backwards=go_backwards,
            stateful=stateful,
            unroll=unroll,  # type: ignore
            **kwargs,
        )
        self.input_spec = [InputSpec(ndim=3)]
        self.parallelization_factor = parallelization_factor
        self._set_unroll()

    def _compute_ebops(self, shape):
        cell_shape = (1, *shape[1:])
        ebops = self.cell._compute_ebops(cell_shape) * self.parallelization_factor
        return ebops

    def get_config(self):  # type: ignore
        base_conf = super().get_config()
        conf = {
            'iq_conf': self.cell.iq.config if self.enable_iq else None,
            'sq_conf': self.cell.sq.config if self.cell.enable_sq else None,
            'kq_conf': self.cell.kq.config,
            'rkq_conf': self.cell.rkq.config,
            'bq_conf': self.cell.bq.config if self.cell.use_bias else None,
            'oq_conf': self.cell.oq.config if self.enable_oq else None,
            'paq_conf': self.cell.paq.config,
        }
        return {**base_conf, **conf}


QLayerBase.register(QSimpleRNN)



from math import prod

from keras import constraints, initializers, ops, regularizers
from keras.layers import Dense

from ...quantizer import Quantizer
from ...quantizer.config import QuantizerConfig
from ...utils.misc import gather_vars_to_kwargs
from .base import QLayerBaseSingleInput


class QDense(QLayerBaseSingleInput, Dense):
    def __init__(
        self,
        units,
        activation=None,
        use_bias=True,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        kq_conf: None | QuantizerConfig = None,
        iq_conf: None | QuantizerConfig = None,
        bq_conf: None | QuantizerConfig = None,
        parallelization_factor=-1,
        **kwargs,
    ):
        kwargs = gather_vars_to_kwargs('self|kq_conf|bq_conf|parallelization_factor')
        super().__init__(lora_rank=None, **kwargs)

        kq_conf = kq_conf or QuantizerConfig('default', 'weight')
        self._kq = Quantizer(kq_conf, name=f'{self.name}_kq')
        bq_conf = bq_conf or QuantizerConfig('default', 'bias')
        self._bq = Quantizer(bq_conf, name=f'{self.name}_bq') if use_bias else None
        self.parallelization_factor = parallelization_factor

    @property
    def kq(self):
        return self._kq

    @property
    def bq(self):
        return self._bq

    def build(self, input_shape):
        super().build(input_shape)
        n_parallel = prod(input_shape[1:-1])
        self.n_parallel = n_parallel
        self.parallelization_factor = self.parallelization_factor if self.parallelization_factor > 0 else n_parallel

        self.kq.build(ops.shape(self._kernel))
        if self.bias is not None:
            assert self.bq is not None
            self.bq.build(ops.shape(self.bias))

    def call(self, inputs, training=None):
        if self.enable_iq:
            inputs = self.iq(inputs, training=training)
        x = ops.matmul(inputs, self.qkernel)
        if self.bias is not None:
            x = ops.add(x, self.qbias)
        if self.activation is not None:
            x = self.activation(x)
        return x

    def _compute_ebops(self, shape):
        bw_inp = self.iq.bits_(shape)
        bw_ker = self.kq.bits_(ops.shape(self.kernel))
        ebops = ops.sum(ops.matmul(bw_inp, bw_ker))
        ebops = ebops * self.n_parallel / self.parallelization_factor  # type: ignore
        if self.bq is not None:
            bw_bias = self.bq.bits_(ops.shape(self.bias))
            size = ops.cast(ops.prod(shape), self.dtype)
            ebops = ebops + ops.mean(bw_bias) * size  # type: ignore

        return ebops

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                'kq_conf': self.kq.config,
                'bq_conf': self.bq.config if self.bq is not None else None,
            }
        )
        return config

    @property
    def qkernel(self):
        return self.kq(self._kernel)

    @property
    def qbias(self):
        if self.bias is None:
            return None
        assert self.bq is not None
        return self.bq(self.bias)


class QBatchNormDense(QDense):
    def __init__(
        self,
        units,
        activation=None,
        use_bias=True,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        scale=True,
        momentum=0.99,
        epsilon=1e-3,
        bn_gamma_initializer='ones',
        moving_mean_initializer='zeros',
        moving_variance_initializer='ones',
        bn_gamma_regularizer=None,
        bn_gamma_constraint=None,
        synchronized=False,
        kq_conf: None | QuantizerConfig = None,
        iq_conf: None | QuantizerConfig = None,
        bq_conf: None | QuantizerConfig = None,
        **kwargs,
    ):
        assert use_bias, 'BatchNormDense must have `use_bias=True`'
        super().__init__(
            units=units,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            kq_conf=kq_conf,
            iq_conf=iq_conf,
            bq_conf=bq_conf,
            **kwargs,
        )

        self.scale = scale
        self.momentum = float(momentum)
        self.bn_gamma_initializer = initializers.get(bn_gamma_initializer)
        self.moving_mean_initializer = initializers.get(moving_mean_initializer)
        self.moving_variance_initializer = initializers.get(moving_variance_initializer)
        self.bn_gamma_regularizer = regularizers.get(bn_gamma_regularizer)
        self.bn_gamma_constraint = constraints.get(bn_gamma_constraint)
        self.epsilon = epsilon
        self.synchronized = synchronized

    def build(self, input_shape):
        super().build(input_shape)
        shape = (input_shape[-1],)
        self.reduction_axis = tuple(range(len(input_shape) - 1))
        if self.scale:
            self.bn_gamma = self.add_weight(
                shape=shape,
                name='bn_gamma',
                initializer=self.bn_gamma_initializer,
                regularizer=self.bn_gamma_regularizer,
                constraint=self.bn_gamma_constraint,
                trainable=True,
                autocast=False,
            )

        self.moving_mean = self.add_weight(
            shape=shape,
            name='moving_mean',
            initializer=self.moving_mean_initializer,
            trainable=False,
            autocast=False,
        )
        self.moving_variance = self.add_weight(
            shape=shape,
            name='moving_variance',
            initializer=self.moving_variance_initializer,
            trainable=False,
            autocast=False,
        )

    def get_fused_qkernel_and_qbias(self, mean, var):
        assert self.bias is not None
        assert self.bq is not None
        if self.scale:
            bn_gamma = self.bn_gamma
        else:
            bn_gamma = 1
        scaler = bn_gamma / ops.sqrt(var + self.epsilon)  # type: ignore
        kernel = ops.cast(self.kernel, self.kernel.dtype)
        fused_qkernel = self.kq(scaler[:, None] * kernel)  # type: ignore

        offset = -ops.dot(mean, kernel)
        fused_qbias = self.bq(self.bias + offset)

        return fused_qkernel, fused_qbias

    def call(self, inputs, training=None):
        if self.enable_iq:
            inputs = self.iq(inputs, training=training)

        if training and self.trainable:
            mean, var = ops.moments(inputs, self.reduction_axis, keepdims=False, synchronized=self.synchronized)  # type: ignore
            self.moving_mean.assign(
                self.moving_mean * self.momentum + mean * (1.0 - self.momentum),
            )
            self.moving_variance.assign(
                self.moving_variance * self.momentum + var * (1.0 - self.momentum),
            )
            # _var, _mean = ops.cast(var, self.moving_variance.dtype), ops.cast(mean, self.moving_mean.dtype)
            # var = ops.stop_gradient(_var - var) + var
            # mean = ops.stop_gradient(_mean - mean) + mean
        else:
            var, mean = self.moving_variance, self.moving_mean

        qkernel, qbias = self.get_fused_qkernel_and_qbias(mean, var)
        x = ops.matmul(inputs, qkernel)
        x = ops.add(x, qbias)
        if self.activation is not None:
            x = self.activation(x)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                'kq_conf': self.kq.config,
                'bq_conf': self.bq.config if self.bq is not None else None,
                'scale': self.scale,
                'momentum': self.momentum,
                'epsilon': self.epsilon,
                'bn_gamma_initializer': self.bn_gamma_initializer,
                'moving_mean_initializer': self.moving_mean_initializer,
                'moving_variance_initializer': self.moving_variance_initializer,
                'bn_gamma_regularizer': self.bn_gamma_regularizer,
                'bn_gamma_constraint': self.bn_gamma_constraint,
                'synchronized': self.synchronized,
            }
        )
        return config

    @property
    def qkernel(self):
        return self.get_fused_qkernel_and_qbias(self.moving_mean, self.moving_variance)[0]

    @property
    def qbias(self):
        return self.get_fused_qkernel_and_qbias(self.moving_mean, self.moving_variance)[1]



from keras.layers import BatchNormalization
from keras.src import backend, ops
from keras.src.backend import standardize_dtype

from ..quantizer import Quantizer
from ..quantizer.config import QuantizerConfig
from ..utils.misc import gather_vars_to_kwargs
from .core.base import QLayerBaseSingleInput


class QBatchNormalization(QLayerBaseSingleInput, BatchNormalization):
    @property
    def beta(self):
        if self._beta is None:
            return backend.convert_to_tensor(0)
        return ops.cast(self._beta, ops.dtype(self._beta))

    @beta.setter
    def beta(self, value):
        """Workaround for setting the batch normalization beta during building."""
        if self.built:
            raise AttributeError('Cannot set beta after build.')
        self.bn_beta = value

    @property
    def gamma(self):
        raise AttributeError('Use `bn_gamma` instead of `gamma`.')

    @gamma.setter
    def gamma(self, value):
        """Workaround for setting the batch normalization gamma during building."""
        if self.built:
            raise AttributeError('Cannot set gamma after build.')
        self.bn_gamma = value

    def __init__(
        self,
        axis=-1,
        momentum=0.99,
        epsilon=1e-3,
        center=True,
        scale=True,
        beta_initializer='zeros',
        gamma_initializer='ones',
        moving_mean_initializer='zeros',
        moving_variance_initializer='ones',
        beta_regularizer=None,
        gamma_regularizer=None,
        beta_constraint=None,
        gamma_constraint=None,
        synchronized=False,
        kq_conf: None | QuantizerConfig = None,
        iq_conf: None | QuantizerConfig = None,
        bq_conf: None | QuantizerConfig = None,
        **kwargs,
    ):
        kwargs = gather_vars_to_kwargs('self|kq_conf|bq_conf')
        super().__init__(**kwargs)

        kq_conf = kq_conf or QuantizerConfig('default', 'weight')
        self._kq = Quantizer(kq_conf, name=f'{self.name}_kq')
        bq_conf = bq_conf or QuantizerConfig('default', 'bias')
        self._bq = Quantizer(bq_conf, name=f'{self.name}_bq')

    @property
    def kq(self):
        return self._kq

    @property
    def bq(self):
        return self._bq

    def build(self, input_shape):
        super().build(input_shape)
        if self.bq is not None:
            self.bq.build(ops.shape(self.moving_mean))
        if self.kq is not None:
            self.kq.build(ops.shape(self.moving_variance))
        shape = [1] * len(input_shape)
        shape[self.axis] = input_shape[self.axis]
        self._shape = tuple(shape)

        # using name / path or _name / _path depends on the keras version.
        if self.bn_beta is not None:
            if hasattr(self.bn_beta, '_name'):
                self.bn_beta._name = 'bn_beta'
                self.bn_beta._path = self.bn_beta._path.replace('/beta', '/bn_beta')
            else:
                self.bn_beta.name = 'bn_beta'
                self.bn_beta.path = self.bn_beta.path.replace('/beta', '/bn_beta')

        if self.bn_gamma is not None:
            if hasattr(self.bn_gamma, '_name'):
                self.bn_gamma._name = 'bn_gamma'
                self.bn_gamma._path = self.bn_gamma._path.replace('/gamma', '/bn_gamma')
            else:
                self.bn_gamma.name = 'bn_gamma'
                self.bn_gamma.path = self.bn_gamma.path.replace('/gamma', '/bn_gamma')

    def _scaler_and_offset(self):
        mean = ops.cast(self.moving_mean, self.dtype)
        variance = ops.cast(self.moving_variance, self.dtype)

        if self.scale:
            bn_gamma = ops.cast(self.bn_gamma, self.dtype)
        else:
            bn_gamma = 1

        if self.center:
            bn_beta = ops.cast(self.bn_beta, self.dtype)
        else:
            bn_beta = 1

        scale = bn_gamma / ops.sqrt(variance + self.epsilon)  # type: ignore
        offset = bn_beta - mean * scale  # type: ignore

        return scale, offset

    @property
    def qscaler_and_qoffset(self):
        scale, offset = self._scaler_and_offset()
        return self.kq(scale, training=False), self.bq(offset, training=False)

    def _scaler_and_offset_train(self, qinputs, mask):
        moving_mean = ops.cast(self.moving_mean, ops.dtype(qinputs))
        moving_variance = ops.cast(self.moving_variance, ops.dtype(qinputs))

        mean, variance = self._moments(qinputs, mask)  # type: ignore
        self.moving_mean.assign(  # type: ignore
            moving_mean * self.momentum + mean * (1.0 - self.momentum),  # type: ignore
        )
        self.moving_variance.assign(  # type: ignore
            moving_variance * self.momentum + variance * (1.0 - self.momentum),  # type: ignore
        )

        if self.scale:
            bn_gamma = ops.cast(self.bn_gamma, ops.dtype(qinputs))
        else:
            bn_gamma = 1

        if self.center:
            bn_beta = ops.cast(self.bn_beta, ops.dtype(qinputs))
        else:
            bn_beta = 1

        scale = bn_gamma / ops.sqrt(variance + self.epsilon)  # type: ignore
        offset = bn_beta - mean * scale  # type: ignore

        return scale, offset

    def call(self, inputs, training=None, mask=None):
        # Check if the mask has one less dimension than the inputs.
        if mask is not None:
            if len(mask.shape) != len(inputs.shape) - 1:
                # Raise a value error
                raise ValueError(
                    'The mask provided should be one dimension less '
                    'than the inputs. Received: '
                    f'mask.shape={mask.shape}, inputs.shape={inputs.shape}',
                )

        input_dtype = standardize_dtype(inputs.dtype)
        if input_dtype in ('float16', 'bfloat16') and training:
            # BN is prone to overflowing for float16/bfloat16 inputs, so we opt
            # out BN for mixed precision.
            inputs = ops.cast(inputs, 'float32')

        if self.enable_iq:
            inputs = self.iq(inputs, training=training)

        if training:
            scale, offset = self._scaler_and_offset_train(inputs, mask)
        else:
            scale, offset = self._scaler_and_offset()

        shape = self._shape

        qscale = self.kq(scale, training=training)
        qoffset = self.bq(offset, training=training)

        qscale = ops.reshape(qscale, shape)
        qoffset = ops.reshape(qoffset, shape)

        outputs = inputs * qscale + qoffset  # type: ignore

        if input_dtype in ('float16', 'bfloat16'):
            outputs = ops.cast(outputs, input_dtype)
        return outputs

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                'kq_conf': self.kq.config,
                'bq_conf': self.bq.config,
            }
        )
        return config

    def _compute_ebops(self, shape):
        bw_inp = self.iq.bits_(shape)
        bw_ker = ops.reshape(self.kq.bits_(self.moving_mean.shape), self._shape)  # type: ignore
        bw_bias = ops.reshape(self.bq.bits_(self.moving_mean.shape), self._shape)  # type: ignore
        size = ops.cast(ops.prod(shape), self.dtype)
        ebops = ops.sum(bw_inp * bw_ker) + ops.mean(bw_bias) * size  # type: ignore
        return ebops
